/* This file is part of the Palabos library.
 * Copyright (C) 2009, 2010 Jonas Latt
 * E-mail contact: jonas@lbmethod.org
 * The most recent release of Palabos can be downloaded at
 * <http://www.lbmethod.org/palabos/>
 *
 * The library Palabos is free software: you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * The library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

/** \file
 * Helper functions for domain initialization -- header file.
 */
#ifndef SEDIMENTATION_EROSION_FUNCTIONAL_3D_HH
#define SEDIMENTATION_EROSION_FUNCTIONAL_3D_HH

#include "sedimentationErosionFunctional3D.h"
#include "core/plbDebug.h"
#include "core/util.h"
#include "core/blockStatistics.h"
#include "latticeBoltzmann/momentTemplates.h"
#include "latticeBoltzmann/geometricOperationTemplates.h"
#include "finiteDifference/fdStencils1D.h"
#include "atomicBlock/atomicBlock3D.h"
#include "atomicBlock/blockLattice3D.h"
#include "atomicBlock/dataField3D.h"
//#include "thrombosisLattices3D.h"
#include "nodeNatureTemplates.h"
#include <cmath>
#include <limits>

namespace plb
{


///////////////  Erosion Process   ////////////////////////////////////

template< typename T1, template<class U> class Descriptor1, typename T2, template<class U> class Descriptor2, typename T3 >
ErosionFromWssFunctional3D<T1,Descriptor1,T2,Descriptor2,T3>::
    ErosionFromWssFunctional3D(T1 stressThreshold_, T1 criticalTime_, T1 criticalNumber_, T1 prob_, T1 omegaNS_, T1 cSmago_, Array<T1,3> gravity_) :
        stressThreshold(stressThreshold_),
        criticalTime(criticalTime_),
        criticalNumber(criticalNumber_),
        prob(prob_),
        omegaNS(omegaNS_),
        cSmago(cSmago_),
        gravity(gravity_)
{  }

template<typename T1, template<class U> class Descriptor1, typename T2, template<class U> class Descriptor2, typename T3>
void ErosionFromWssFunctional3D<T1,Descriptor1,T2,Descriptor2,T3>::
        processGenericBlocks(Box3D domain, std::vector<AtomicBlock3D*> atomicBlocks)
{
    enum { nodeNature  = Descriptor2<T2>::ExternalField::nodeNatureBeginsAt,
           timeErosion = Descriptor2<T2>::ExternalField::timeErosionBeginsAt,
           externalForce = Descriptor2<T2>::ExternalField::forceBeginsAt,
    };

    using namespace freeSurfaceFlag;


    BlockLattice3D<T1,Descriptor1>& particles =
            *dynamic_cast<BlockLattice3D<T1,Descriptor1> *>(atomicBlocks[0]);

    BlockLattice3D<T2,Descriptor2>& lattice =
            *dynamic_cast<BlockLattice3D<T2,Descriptor2> *>(atomicBlocks[1]);

    ScalarField3D<T3>& wss =
            *dynamic_cast<ScalarField3D<T3> *>(atomicBlocks[2]);

    ScalarField3D<int>& flag =
            *dynamic_cast<ScalarField3D<int> *>(atomicBlocks[3]);

    Dot3D offset = computeRelativeDisplacement(particles, lattice);
    Dot3D offWss = computeRelativeDisplacement(particles, wss);
    Dot3D offsetFlag = computeRelativeDisplacement(particles, flag);



    for (plint iX=domain.x0; iX<=domain.x1; ++iX) {
        plint oX = iX + offset.x;
        plint wX = iX + offWss.x;
	    plint flagX = iX + offsetFlag.x;
        for (plint iY=domain.y0; iY<=domain.y1; ++iY) {
            plint oY = iY + offset.y;
            plint wY = iY + offWss.y;
	        plint flagY = iY + offsetFlag.y;
            for (plint iZ=domain.z0; iZ<=domain.z1; ++iZ) {
                plint oZ = iZ + offset.z;
                plint wZ = iZ + offWss.z;
		        plint flagZ = iZ + offsetFlag.z;

                if (*(lattice.get(oX,oY,oZ).getExternal(nodeNature)) == NodeNature::wall &&
		            flag.get(flagX,flagY,flagZ) == wall )
                {

                    T1 numberParticles = particles.get(iX,iY,iZ).computeDensity();

                    // now we should get the wss acting on this node
                    // better way, probably, is to make an average over its neighs (easier in terms of position recon.)

                    T1 averageWss = T1(0);
                    Array<T1,3> averageVelocity ( T1(0), T1(0), T1(0));
                    plint counterOfCoolneighs = 0;

                    bool maskOfErosion = bool(*particles.get(iX,iY,iZ).getExternal(maskOffset));
                    // I enter the if loop if 1. I'm a nextToWall lattice
                    // if my advDiffusion dynamic is not a bounceBack yet.
                    // What I have to introduce is the concept of liquidFraction
                    // If the the liquidFraction is in between 0-1 then I do have
                    // to perform a reaction Step, otherwise not needed.

                    // N.B. liquid fraction is different from one (completely fluid lattice) only on nextToWall
                    // nodes.

                    if( maskOfErosion )
                    {
                        for (plint iPop = 1; iPop < descriptors::D3Q27Descriptor<T1>::q; ++iPop)
                        {

                            plint nextX = oX + descriptors::D3Q27Descriptor<T1>::c[iPop][0];
                            plint nextY = oY + descriptors::D3Q27Descriptor<T1>::c[iPop][1];
                            plint nextZ = oZ + descriptors::D3Q27Descriptor<T1>::c[iPop][2];

                            plint nextX_f = flagX + descriptors::D3Q27Descriptor<T1>::c[iPop][0];
                            plint nextY_f = flagY + descriptors::D3Q27Descriptor<T1>::c[iPop][1];
                            plint nextZ_f = flagZ + descriptors::D3Q27Descriptor<T1>::c[iPop][2];

                            bool a = *(lattice.get(nextX,nextY,nextZ).getExternal(nodeNature)) == NodeNature::nextToWall;
                            bool b = *(lattice.get(nextX,nextY,nextZ).getExternal(nodeNature)) == NodeNature::fluid;

                            bool c = flag.get(nextX_f, nextY_f, nextZ_f) == fluid;
                            bool d = flag.get(nextX_f, nextY_f, nextZ_f) == interface;
                            // is the check on flag needed?
                            /*
                            if ( (*(lattice.get(nextX,nextY,nextZ).getExternal(nodeNature)) == NodeNature::nextToWall ||
                              *(lattice.get(nextX,nextY,nextZ).getExternal(nodeNature)) == NodeNature::fluid ) &&
                              (flag.get(nextX_f,nextY_f,nextZ_f) == fluid || flag.get(nextX_f,nextY_f,nextZ_f) == interface))
                            */
                            if( (a||b) && (c||d))
                            {

                                plint nextX_wss = wX + descriptors::D3Q27Descriptor<T1>::c[iPop][0];
                                plint nextY_wss = wY + descriptors::D3Q27Descriptor<T1>::c[iPop][1];
                                plint nextZ_wss = wZ + descriptors::D3Q27Descriptor<T1>::c[iPop][2];

                                counterOfCoolneighs +=1;

                                Array<T2,3> localVelocity;

                                lattice.get(nextX,nextY,nextZ).computeVelocity(localVelocity);
                                averageVelocity[0] +=  localVelocity[0];
                                averageVelocity[1] +=  localVelocity[1];
                                averageVelocity[2] +=  localVelocity[2];

                                averageWss += fabs(wss.get(nextX_wss,nextY_wss,nextZ_wss));

                            }

                        }
                        vvvvaverageWss /= counterOfCoolneighs;
                        averageVelocity /= counterOfCoolneighs;

                        // well ..now I have the wallShearStress for a BB node
                        //
                        if (averageWss >= stressThreshold)
                        {
                            *(particles.get(oX,oY,oZ).getExternal(fractionAtTimeT)) = kappa_wss_pos*particles.get(oX,oY,oZ).computeDensity();
                        }
                        
                        else
                        {
                            *(particles.get(oX,oY,oZ).getExternal(fractionAtTimeT)) = kappa_wss_neg*particles.get(oX,oY,oZ).computeDensity();
                        }

                    } // endif mask erosion
                    else
                    {
                        *(particles.get(oX,oY,oZ).getExternal(fractionAtTimeT)) = T1(0); // is this assumption Correct?
                    }

                }
            }
        }
    }
}

template<typename T1, template<class U> class Descriptor1, typename T2, template<class U> class Descriptor2, typename T3>
ErosionFromWssFunctional3D<T1,Descriptor1,T2,Descriptor2,T3>*
    ErosionFromWssFunctional3D<T1,Descriptor1,T2,Descriptor2,T3>::clone() const
{
    return new ErosionFromWssFunctional3D<T1,Descriptor1,T2,Descriptor2,T3>(*this);
}

template<typename T1, template<class U> class Descriptor1, typename T2, template<class U> class Descriptor2, typename T3>
void ErosionFromWssFunctional3D<T1,Descriptor1,T2,Descriptor2,T3>::getTypeOfModification (
        std::vector<modif::ModifT>& modified ) const
{
    modified[0] = modif::nothing;
    // Full dynamics object must be recreated, because this data processor
    //   re-attributes a new dynamics and acts on the bulk only.
    modified[1] = modif::dataStructure;
    modified[2] = modif::nothing;
    modified[3] = modif::nothing;
}


#endif  // DATA_ANALYSIS_FUNCTIONAL_3D_HH
