/* This file is part of the Palabos library.
 * Copyright (C) 2009, 2010 Jonas Latt
 * E-mail contact: jonas@lbmethod.org
 * The most recent release of Palabos can be downloaded at 
 * <http://www.lbmethod.org/palabos/>
 *
 * The library Palabos is free software: you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * The library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

/** \file
 * Helper functions for domain initialization -- header file.
 */
#ifndef THROMBOSIS_FUNCTIONAL_3D_H
#define THROMBOSIS_FUNCTIONAL_3D_H

#include "core/globalDefs.h"
#include "atomicBlock/dataProcessingFunctional3D.h"
#include "atomicBlock/reductiveDataProcessingFunctional3D.h"
#include "latticeBoltzmann/geometricOperationTemplates.h"

namespace plb {
    
template<typename T1, template<class U> class Descriptor1, typename T2, template<class U> class Descriptor2>
class CopyVelocityToParticlesFunctional3D : public BoxProcessingFunctional3D_LL<T1,Descriptor1,T2,Descriptor2>
{
public:
    CopyVelocityToParticlesFunctional3D(Array<T1,3> &settlingVelocityVector_ );
    virtual void process(Box3D domain, BlockLattice3D<T1,Descriptor1> &lattice, 
                                       BlockLattice3D<T2,Descriptor2> &particles);
    virtual CopyVelocityToParticlesFunctional3D<T1,Descriptor1,T2,Descriptor2>* clone() const;
    virtual void getTypeOfModification(std::vector<modif::ModifT>& modified) const;
private:
    Array<T1,3> &settlingVelocityVector;
	
};

template<typename T1, template<class U> class Descriptor1, typename T2, template<class U> class Descriptor2>
class MakeThrombosisToParticlesFunctional3D : public BoxProcessingFunctional3D_LL<T1,Descriptor1,T2,Descriptor2>
{
public:
    MakeThrombosisToParticlesFunctional3D(T1 omegaAD_, T1 iniParticles_, Dynamics<T2,Descriptor2>* dynamicsParticles_);
    virtual void process(Box3D domain, BlockLattice3D<T1,Descriptor1> &lattice, 
                         BlockLattice3D<T2,Descriptor2> &particles);
    virtual MakeThrombosisToParticlesFunctional3D<T1,Descriptor1,T2,Descriptor2>* clone() const;
    virtual void getTypeOfModification(std::vector<modif::ModifT>& modified) const;
private:
    T1 omegaAD, iniParticles;
    Dynamics<T2,Descriptor2>* dynamicsParticles;
    
    
};


template<typename T1, template<class U> class Descriptor1, typename T2, template<class U> class Descriptor2, typename T3>
class MakeThrombosisFromWssFunctional3D : public BoxProcessingFunctional3D
{
public :
    MakeThrombosisFromWssFunctional3D(T1 stressThreshold_, T1 criticalTime_, T1 criticalNumber_, T1 prob_);
    virtual void processGenericBlocks(Box3D domain, std::vector<AtomicBlock3D*> atomicBlocks);

    virtual MakeThrombosisFromWssFunctional3D<T1,Descriptor1,T2,Descriptor2,T3>* clone() const;
    virtual void getTypeOfModification(std::vector<modif::ModifT>& modified) const;
private :
    T1 stressThreshold, criticalTime, criticalNumber, prob;
};


template<typename T1, template<class U> class Descriptor1, typename T2, template<class U> class Descriptor2, typename T3>
class ErosionFromWssFunctional3D : public BoxProcessingFunctional3D
{
public :
    ErosionFromWssFunctional3D(T1 stressThreshold_, T1 criticalTime_, T1 criticalNumber_, T1 prob_, 
                              T1 omegaNS, T1 cSmago_,
			      Array<T1,3> gravity_, 
                              Dynamics<T1,Descriptor1>* dynamics_,
                              Dynamics<T2,Descriptor2>* dynamicsParticles_ );
    virtual void processGenericBlocks(Box3D domain, std::vector<AtomicBlock3D*> atomicBlocks);

    virtual ErosionFromWssFunctional3D<T1,Descriptor1,T2,Descriptor2,T3>* clone() const;
    virtual void getTypeOfModification(std::vector<modif::ModifT>& modified) const;
private :
    T1 stressThreshold, criticalTime, criticalNumber, prob, omegaNS, cSmago;
    Array<T1,3> gravity;
    Dynamics<T1,Descriptor1>* dynamics;
    Dynamics<T2,Descriptor2>* dynamicsParticles;
};




template<typename T, template<class U> class Descriptor>
class SetParticleNumber3D : public BoxProcessingFunctional3D_L<T,Descriptor>
{
public :
    SetParticleNumber3D(T number_) : number(number_)
    { }
    
    virtual void process(Box3D domain, BlockLattice3D<T,Descriptor>& lattice);

    virtual SetParticleNumber3D<T,Descriptor>* clone() const;
    virtual void getTypeOfModification(std::vector<modif::ModifT>& modified) const;
private :
    T number;
};

template<typename T, template<class U> class Descriptor, int direction, int orientation>
class ZeroGradientParticleNumber3D : public BoxProcessingFunctional3D_L<T,Descriptor>
{
public :
    virtual void process(Box3D domain, BlockLattice3D<T,Descriptor>& lattice);

    virtual ZeroGradientParticleNumber3D<T,Descriptor,direction,orientation>* clone() const;
    virtual void getTypeOfModification(std::vector<modif::ModifT>& modified) const;
};



template<typename T1, template<class U> class Descriptor1, typename T2, template<class U> class Descriptor2>
class ChangeDynamicForAdvectionDiffusionDynamic3D : public BoxProcessingFunctional3D{
public:
    ChangeDynamicForAdvectionDiffusionDynamic3D(T2 omegaAD_, Dynamics<T2,Descriptor2>* dynamics_);
    virtual void processGenericBlocks(Box3D domain, std::vector<AtomicBlock3D*> atomicBlocks);
    virtual ChangeDynamicForAdvectionDiffusionDynamic3D<T1,Descriptor1,T2,Descriptor2>* clone() const;
     // (1) lattice, (2) particles, (3) flag-status
    virtual void getTypeOfModification(std::vector<modif::ModifT>& modified) const;
private:
    T2 omegaAD;  
    Dynamics<T2,Descriptor2>* dynamics;
};


template<typename T1, template<class U> class Descriptor1, typename T2, template<class U> class Descriptor2>
class CopyUnknownFromNeighsFunctional3D : public BoxProcessingFunctional3D{
public:
    virtual void processGenericBlocks(Box3D domain, std::vector<AtomicBlock3D*> atomicBlocks);
    virtual CopyUnknownFromNeighsFunctional3D<T1,Descriptor1,T2,Descriptor2>* clone() const;
     // (1) fluid, (2) mass-fraction, (3) flag-status
    virtual void getTypeOfModification(std::vector<modif::ModifT>& modified) const;
};


template<typename T1, template<class U> class Descriptor1, typename T2, template<class U> class Descriptor2>
class TestTurbulentAdvectionDiffusionFunctional3D : public BoxProcessingFunctional3D_LL<T1,Descriptor1,T2,Descriptor2>
{
public:
    virtual void process(Box3D domain, BlockLattice3D<T1,Descriptor1> &lattice, 
                         BlockLattice3D<T2,Descriptor2> &particles);
    virtual TestTurbulentAdvectionDiffusionFunctional3D<T1,Descriptor1,T2,Descriptor2>* clone() const;
    virtual void getTypeOfModification(std::vector<modif::ModifT>& modified) const;
    
    
};


template<typename T1, template<class U> class Descriptor1>
class CalculateTemporalWssAverage3D : public BoxProcessingFunctional3D
{
public :
    virtual void processGenericBlocks(Box3D domain, std::vector<AtomicBlock3D*> atomicBlocks);
   
    virtual CalculateTemporalWssAverage3D<T1,Descriptor1>* clone() const;
    virtual void getTypeOfModification(std::vector<modif::ModifT>& modified) const;

};


template<typename T1, template<class U> class Descriptor1>
class CopyScalarFromExternalFieldToScalarField : public BoxProcessingFunctional3D_LS<T1,Descriptor1,T1>
{
public :
    virtual void process(Box3D domain, BlockLattice3D<T1,Descriptor1>& lattice, ScalarField3D<T1>& copyTo);
    
    virtual CopyScalarFromExternalFieldToScalarField<T1,Descriptor1>* clone() const;
    virtual void getTypeOfModification(std::vector<modif::ModifT>& modified) const;
};


template<typename T1, template<class U> class Descriptor1>
class CopyStateOfLocalDynamicToScalarField : public BoxProcessingFunctional3D_LS<T1,Descriptor1,T1>
{
public :
    virtual void process(Box3D domain, BlockLattice3D<T1,Descriptor1>& particles, ScalarField3D<T1>& copyTo);
    
    virtual CopyStateOfLocalDynamicToScalarField<T1,Descriptor1>* clone() const;
    virtual void getTypeOfModification(std::vector<modif::ModifT>& modified) const;
};


template<typename T1, template<class U> class Descriptor1>
class ChangeErodableCellStatus3D : public BoxProcessingFunctional3D_LS<T1,Descriptor1,int>
{
public :
    ChangeErodableCellStatus3D(plint erodableWallLimit_);
    virtual void process(Box3D domain, BlockLattice3D<T1,Descriptor1>& lattice, ScalarField3D<int>& flag);
    
    virtual ChangeErodableCellStatus3D<T1,Descriptor1>* clone() const;
    virtual void getTypeOfModification(std::vector<modif::ModifT>& modified) const;
private:
    plint  erodableWallLimit;  
};



}  // namespace plb

#endif  // THROMBOSIS_FUNCTIONAL_3D_H
