/* This file is part of the Palabos library.
 *
 * Copyright (C) 2011-2015 FlowKit Sarl
 * Route d'Oron 2
 * 1010 Lausanne, Switzerland
 * E-mail contact: contact@flowkit.com
 *
 * The most recent release of Palabos can be downloaded at 
 * <http://www.palabos.org/>
 *
 * The library Palabos is free software: you can redistribute it and/or
 * modify it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * The library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef MESH_TO_VOXEL_HH
#define MESH_TO_VOXEL_HH

#include "core/globalDefs.h"
#include "atomicBlock/dataField3D.h"
#include "plbWrapper/block/dataInitializerWrapper3D.h"
#include "plbWrapper/block/dataAnalysisWrapper3D.h"
#include "boundary/meshToVoxel.h"
#include "geometry/triangleHashing.h"
#include "multiBlock/multiBlockGenerator3D.h"
#include "dataProcessors/metaStuffWrapper3D.h"
#include "core/plbTimer.h"
#include "multiBlock/defaultMultiBlockPolicy3D.h"
#include "multiBlock/nonLocalTransfer3D.h"
#include "multiBlock/multiDataProcessorWrapper3D.h"
#include "multiBlock/reductiveMultiDataProcessorWrapper3D.h"
#include "offLattice/voxelizer.h"
#include "geometry/triangleUtil.h"

namespace plb {

template<typename T>
MultiNTensorField3D<int>* meshToVoxel (
        ConnectedTriangleMesh<T>& mesh,
        plint symmetricLayer, plint borderWidth )
{
    Cuboid<T> cuboid = mesh.getBoundingCuboid();
    T dX = cuboid.upperRightCorner[0] - cuboid.lowerLeftCorner[0];
    T dY = cuboid.upperRightCorner[1] - cuboid.lowerLeftCorner[1];
    T dZ = cuboid.upperRightCorner[2] - cuboid.lowerLeftCorner[2];

    // Creation of the multi-scalar field. The +1 is because if the resolution is N,
    //   the number of nodes is N+1.
    plint nx = (plint) dX + 1 + 2*symmetricLayer;
    plint ny = (plint) dY + 1 + 2*symmetricLayer;
    plint nz = (plint) dZ + 1 + 2*symmetricLayer;

    return meshToVoxel(mesh, Box3D(0,nx-1, 0,ny-1, 0,nz-1), borderWidth);
}

template<typename T>
MultiNTensorField3D<int>* meshToVoxel (
        ConnectedTriangleMesh<T>& mesh,
        Box3D const& domain, plint borderWidth )
{
    // As initial seed, a one-cell layer around the outer boundary is tagged
    //   as outside cells.
    plint envelopeWidth = 1;
    plint nDim = 1;
    int iniVal = voxelFlag::outside;
    MultiNTensorField3D<int>* voxelMatrix
        = generateMultiNTensorField3D<int>(domain, nDim, &iniVal, envelopeWidth);
    setToConstant( *voxelMatrix, voxelMatrix->getBoundingBox().enlarge(-1),
                   voxelFlag::undetermined );

    MultiContainerBlock3D hashContainer(*voxelMatrix);
    std::vector<MultiBlock3D*> container_arg;
    container_arg.push_back(&hashContainer);
    applyProcessingFunctional (
            new MeshToHash<T>(mesh),
            hashContainer.getBoundingBox(), container_arg );

    std::vector<MultiBlock3D*> flag_hash_arg;
    flag_hash_arg.push_back(voxelMatrix);
    flag_hash_arg.push_back(&hashContainer);

    voxelMatrix->resetFlags(); // Flags are used internally by MeshToVoxelFunctional3D.
    plint maxIter = 5000;
    plint i=0;
    while (!allFlagsTrue(voxelMatrix) && i<maxIter) {
        applyProcessingFunctional (
                new MeshToVoxelFunctional3D<T>(mesh),
                voxelMatrix->getBoundingBox(), flag_hash_arg );
        ++i;
    }
    if (i==maxIter) {
        pcout << "Warning: Voxelization failed." << std::endl;
    }

    // Add the "innerBorder" and "outerBorder" flags.
    identifyBoundary(*voxelMatrix, voxelMatrix->getBoundingBox().enlarge(-1), borderWidth);

    return voxelMatrix;
}

template<typename T>
MultiNTensorField3D<int>* meshToVoxel (
        ConnectedTriangleMesh<T>& mesh,
        Box3D const& domain, plint borderWidth, Box3D seed )
{
    // As initial seed, a one-cell layer around the outer boundary is tagged
    //   as ouside cells.
    plint envelopeWidth=1;
    plint nDim = 1;
    int iniVal = voxelFlag::undetermined;
    MultiNTensorField3D<int>* voxelMatrix
        = generateMultiNTensorField3D<int>(domain, nDim, &iniVal, envelopeWidth);
    setToConstant(*voxelMatrix, seed, voxelFlag::outside);

    MultiContainerBlock3D hashContainer(*voxelMatrix);
    setupHashing(hashContainer, mesh);

    std::vector<MultiBlock3D*> flag_hash_arg;
    flag_hash_arg.push_back(voxelMatrix);
    flag_hash_arg.push_back(&hashContainer);

    voxelMatrix->resetFlags(); // Flags are used internally by MeshToVoxelFunctional3D.
    plint maxIteration=100;
    plint i=0;
    while (!allFlagsTrue(voxelMatrix) && i<maxIteration) {
        applyProcessingFunctional (
                new MeshToVoxelFunctional3D<T>(mesh),
                voxelMatrix->getBoundingBox(), flag_hash_arg );
        ++i;
    }
    if (i==maxIteration) {
        pcout << "Warning: Voxelization failed." << std::endl;
    }

    // Add the "innerBorder" and "outerBorder" flags.
    identifyBoundary(*voxelMatrix, voxelMatrix->getBoundingBox().enlarge(-1), borderWidth);

    return voxelMatrix;
}

template<typename T>
void reMeshToVoxel (
        ConnectedTriangleMesh<T>& mesh, MultiNTensorField3D<int>& voxels, plint borderWidth )
{
    // We don't know if old boundary cells remain boundary, so we switch them to undetermined.
    applyProcessingFunctional (
            new BoundaryToUndetFunctional<int>(),
            voxels.getBoundingBox(), voxels );
    MultiNTensorField3D<int>* undetVoxels = new MultiNTensorField3D<int>(voxels);
    // Neighboring cells may have become boundary cell, so we enlarge the area of undetermined
    // cells by one cell. Attention: it is important to exclude the outermost boundary of the
    // fluid domain from this operation, because otherwise, envelope cells will be accessed
    // which may default to zero if the domain is not periodic.
    growDomain(voxels, voxels.getBoundingBox().enlarge(-1), voxelFlag::undetermined);

    MultiContainerBlock3D hashContainer(voxels);
    std::vector<MultiBlock3D*> container_arg;
    container_arg.push_back(&hashContainer);
    applyProcessingFunctional (
            new MeshToHash<T>(mesh),
            hashContainer.getBoundingBox(), container_arg );

    std::vector<MultiBlock3D*> flag_hash_arg;
    flag_hash_arg.push_back(&voxels);
    flag_hash_arg.push_back(&hashContainer);

    voxels.resetFlags(); // Flags are used internally by MeshToVoxelFunctional3D.
    plint maxIter = 5000;
    plint i=0;
    while (!allFlagsTrue(&voxels) && i<maxIter) {
        applyProcessingFunctional (
                new MeshToVoxelFunctional3D<T>(mesh),
                voxels.getBoundingBox(), flag_hash_arg );
        ++i;
    }
    if (i==maxIter) {
        pcout << "Warning: Voxelization failed." << std::endl;
    }


    // Add the "innerBorder" and "outerBorder" flags.
    reIdentifyBoundary(voxels, *undetVoxels, voxels.getBoundingBox().enlarge(-1), borderWidth);
    delete undetVoxels;
}



/* ******** MeshToVoxelFunctional3D ************************************* */

template<typename T>
MeshToVoxelFunctional3D<T>::MeshToVoxelFunctional3D (
        ConnectedTriangleMesh<T>& mesh_)
    : mesh(mesh_)
{ }

template<typename T>
bool MeshToVoxelFunctional3D<T>::distanceToSurface (
        AtomicContainerBlock3D& hashContainer,
        Array<T,3> const& point, T& distance, bool& isBehind ) const
{
    T maxDistance = std::sqrt((T)3);
    Array<T,2> xRange(point[0]-maxDistance, point[0]+maxDistance);
    Array<T,2> yRange(point[1]-maxDistance, point[1]+maxDistance);
    Array<T,2> zRange(point[2]-maxDistance, point[2]+maxDistance);
    TriangleHashing<T> triangleHash(hashContainer);
    std::vector<plint> possibleTriangles;
    triangleHash.getTriangles(xRange, yRange, zRange, possibleTriangles);

    T    tmpDistance;
    bool tmpIsBehind;
    bool triangleFound = false;

    for (pluint iPossible=0; iPossible<possibleTriangles.size(); ++iPossible) {
        plint triangleID = possibleTriangles[iPossible];
        typename ConnectedTriangleMesh<T>::PTriangle triangle = mesh.triangle(triangleID);
        distanceToTriangle (
                    point, *triangle, tmpDistance, tmpIsBehind );
        if( !triangleFound || tmpDistance<distance) {
            distance = tmpDistance;
            isBehind = tmpIsBehind;
            triangleFound = true;
        }
    }
    return triangleFound;
}

template<typename T>
bool MeshToVoxelFunctional3D<T>::checkIfFacetsCrossed (
        AtomicContainerBlock3D& hashContainer,
        Array<T,3> const& point1, Array<T,3> const& point2,
        T& distance, plint& whichTriangle )
{
    Array<T,2> xRange (
                 std::min(point1[0], point2[0]),
                 std::max(point1[0], point2[0]) );
    Array<T,2> yRange (
                 std::min(point1[1], point2[1]),
                 std::max(point1[1], point2[1]) );
    Array<T,2> zRange (
                 std::min(point1[2], point2[2]),
                 std::max(point1[2], point2[2]) );
    TriangleHashing<T> triangleHash(hashContainer);
    std::vector<plint> possibleTriangles;
    triangleHash.getTriangles(xRange, yRange, zRange, possibleTriangles);

    int flag = 0; // Check for crossings inside the point1-point2 segment.
    Array<T,3> intersection; // Dummy variable.
    Array<T,3> normal;       // Dummy variable.
    T tmpDistance;           // Dummy variable.

    if (global::counter("voxelizer-debug").getCount()==1) {
        std::cout << "{";
    }
    std::vector<T> crossings;
    for (pluint iPossible=0; iPossible<possibleTriangles.size(); ++iPossible) {
        plint triangleID = possibleTriangles[iPossible];
        typename ConnectedTriangleMesh<T>::PTriangle triangle = mesh.triangle(triangleID);
        if (pointOnTriangle(point1, point2, flag, *triangle, intersection, normal, tmpDistance)==1) {
            if (global::counter("voxelizer-debug").getCount()==1) {
                std::cout << "(" << triangleID << ";" << tmpDistance << ")";
            }
            crossings.push_back(tmpDistance);
            if (crossings.size()==1 || tmpDistance<distance) {
                distance = tmpDistance;
                whichTriangle = triangleID;
            }
        }
    }
    if (global::counter("voxelizer-debug").getCount()==1) {
        std::cout << "}";
    }

    if (crossings.size()==0) {
        return false;
    }
    else {
        bool hasCrossed = true;
        for (pluint iCrossing=1; iCrossing<crossings.size(); ++iCrossing) {
            const T eps1 = std::numeric_limits<double>::epsilon()*1.e4;
            if ( std::fabs(crossings[iCrossing]-crossings[iCrossing-1])>eps1)
            {
                hasCrossed = !hasCrossed;
            }
        }
        return hasCrossed;
    }
}

template<typename T>
bool MeshToVoxelFunctional3D<T>::createVoxelizationRange (
        Box3D const& domain, NTensorField3D<int>& voxels,
        Array<plint,2>& xRange, Array<plint,2>& yRange, Array<plint,2>& zRange )
{
    // The purpose of the first three loops is to locate the eight
    //   corners of the cube. One voxel per corner would be insufficient
    //   because a potential seed is situated differently, depending on
    //   whether it is on the boundary of the multi-block or somewhere inside.
    for (plint dx=0; dx<=+1; ++dx) {
        plint xMin = domain.x0+dx*domain.getNx()-1;
        plint xMax = domain.x0+dx*domain.getNx();
        for (plint dy=0; dy<=+1; ++dy) {
            plint yMin = domain.y0+dy*domain.getNy()-1;
            plint yMax = domain.y0+dy*domain.getNy();
            for (plint dz=0; dz<=+1; ++dz) {
                plint zMin = domain.z0+dz*domain.getNz()-1;
                plint zMax = domain.z0+dz*domain.getNz();

                // Locate a potential seed in one of the corners.
                for (plint iX=xMin; iX<=xMax; ++iX) {
                    for (plint iY=yMin; iY<=yMax; ++iY) {
                        for (plint iZ=zMin; iZ<=zMax; ++iZ) {
                            if (*voxels.get(iX,iY,iZ) != voxelFlag::undetermined) {
                                xRange[0] = domain.x0+dx*(domain.getNx()-1);
                                xRange[1] = domain.x0+(1-dx)*(domain.getNx()-1);
                                yRange[0] = domain.y0+dy*(domain.getNy()-1);
                                yRange[1] = domain.y0+(1-dy)*(domain.getNy()-1);
                                zRange[0] = domain.z0+dz*(domain.getNz()-1);
                                zRange[1] = domain.z0+(1-dz)*(domain.getNz()-1);
                                return true;
                            }
                        }
                    }
                }

            }
        }
    }
    return false;
}

template<typename T>
void MeshToVoxelFunctional3D<T>::printOffender (
        NTensorField3D<int> const& voxels,
        AtomicContainerBlock3D& hashContainer,
        Dot3D pos )
{
    std::set<plint> triangles;
    Dot3D offset = voxels.getLocation();
    Dot3D pos_ = pos+offset;
    std::cout << "Position (" << pos_.x << "," << pos_.y << "," << pos_.z << ")" << std::endl;
    for (plint dx=-1; dx<=+1; ++dx) {
        for (plint dy=-1; dy<=+1; ++dy) {
            for (plint dz=-1; dz<=+1; ++dz) {
                if (!(dx==0 && dy==0 && dz==0)) {
                    Dot3D neigh = pos+offset+Dot3D(dx,dy,dz);
                    int typeOfNeighbor = *voxels.get(pos.x+dx,pos.y+dy,pos.z+dz);
                    if (typeOfNeighbor!=voxelFlag::undetermined) {
                        T distance;
                        plint whichTriangle;
                        Array<T,3> p1(pos_.x,pos_.y,pos_.z);
                        Array<T,3> p2(neigh.x,neigh.y,neigh.z);
                        global::counter("voxelizer-debug").increment(1);
                        bool crossed = checkIfFacetsCrossed (
                                    hashContainer, p1, p2, distance, whichTriangle);
                        global::counter("voxelizer-debug").reset();
                        std::cout << "Neighbor ("
                                  << dx << "," << dy << "," << dz
                                  << "); is "
                                  << (voxelFlag::insideFlag(typeOfNeighbor) ? "inside" : "outside");
                        if (crossed) {
                            triangles.insert(whichTriangle);
                            std::cout 
                                  << " inters. at distance " << distance
                                  << " with triangle " << whichTriangle << std::endl;
                        }
                        else {
                            std::cout << " no inters." << std::endl;
                        }
                    }
                }
            }
        }
    }
    std::set<plint>::iterator it = triangles.begin();
    for (; it!=triangles.end(); ++it) {
        std::cout << "Triangle " << *it << " [" << std::flush;
        typename ConnectedTriangleMesh<T>::PTriangle triangle = mesh.triangle(*it);
        Array<T,3> p0 = (*triangle)[0];
        Array<T,3> p1 = (*triangle)[1];
        Array<T,3> p2 = (*triangle)[2];
        std::cout << p0[0] << " " << p1[0] << " " << p2[0] << " " << p0[0] << "], ["
                  << p0[1] << " " << p1[1] << " " << p2[1] << " " << p0[1] << "], ["
                  << p0[2] << " " << p1[2] << " " << p2[2] << " " << p0[2] << "]" << std::endl;
    }
}

template<typename T>
bool MeshToVoxelFunctional3D<T>::voxelizeFromNeighbor (
        NTensorField3D<int> const& voxels,
        AtomicContainerBlock3D& hashContainer,
        Dot3D pos, Dot3D neighbor, int& voxelType )
{
    int verificationLevel = 0;
    Dot3D offset = voxels.getLocation();
    int typeOfNeighbor = *voxels.get(neighbor.x,neighbor.y,neighbor.z);
    if (typeOfNeighbor==voxelFlag::undetermined) {
        return true;
    }
    // If there is no verification and the voxel has already been voxelized,
    //   it is not being re-voxelized here.
    if (verificationLevel==0) {
        if (voxelType!=voxelFlag::undetermined) {
            return true;
        }
    }
    Dot3D pos_ = pos+offset;
    Dot3D neighbor_ = neighbor+offset;
    Array<T,3> point1((T)pos_.x, (T)pos_.y, (T)pos_.z);
    Array<T,3> point2((T)neighbor_.x, (T)neighbor_.y, (T)neighbor_.z);
    int newVoxelType = voxelFlag::undetermined;
    T distance1, distance2, distance3, distance4;
    bool isBehind1, isBehind2;
    plint whichTriangle1, whichTriangle2;
    if (checkIfFacetsCrossed(hashContainer, point1, point2, distance1, whichTriangle1)) {
        newVoxelType = voxelFlag::invert(typeOfNeighbor);
        // Additional consistency checks only at the ultimate level of verification.
        if (verificationLevel==2) {
            PLB_ASSERT( distance1 < std::sqrt((T)3)+(T)0.0001 );
#ifdef PLB_DEBUG
            bool ok = checkIfFacetsCrossed(hashContainer, point2, point1, distance2, whichTriangle2);
#else
            (void) checkIfFacetsCrossed(hashContainer, point2, point1, distance2, whichTriangle2);
#endif
            PLB_ASSERT( ok );
            PLB_ASSERT( distance2 < std::sqrt((T)3)+(T)0.0001 );

#ifdef PLB_DEBUG
            bool ok1 = distanceToSurface( hashContainer, point1, distance3, isBehind1 );
#else
            (void) distanceToSurface( hashContainer, point1, distance3, isBehind1 );
#endif

            PLB_ASSERT( ok1 );
            PLB_ASSERT( distance1 < std::sqrt((T)3)+(T)0.0001 );
            // Attention: At this moment, the following consistency check fails sometimes,
            //   god knows why. It might be that there is a bug in the method
            //   mesh.distanceToSurface.
            PLB_ASSERT( (voxelFlag::insideFlag(newVoxelType) && isBehind1) ||
                        (voxelFlag::outsideFlag(newVoxelType) && !isBehind1) );

#ifdef PLB_DEBUG
            bool ok2 = distanceToSurface( hashContainer, point2, distance4, isBehind2 );
#else
            (void) distanceToSurface( hashContainer, point2, distance4, isBehind2 );
#endif
            PLB_ASSERT( ok2 );
            PLB_ASSERT( distance2 < std::sqrt((T)3)+(T)0.0001 );
            PLB_ASSERT ( (voxelFlag::insideFlag(typeOfNeighbor) && isBehind2) ||
                         (voxelFlag::outsideFlag(typeOfNeighbor) && !isBehind2) );
        }
    }
    else {
        newVoxelType = typeOfNeighbor;
    }
    int oldVoxelType = voxelType;
    voxelType = newVoxelType;
    if (oldVoxelType == voxelFlag::undetermined) {
        return true;
    }
    else {
        return oldVoxelType == newVoxelType;
    }
}

template<typename T>
void MeshToVoxelFunctional3D<T>::processGenericBlocks (
        Box3D domain, std::vector<AtomicBlock3D*> blocks )
{
    PLB_PRECONDITION( blocks.size()==2 );
    NTensorField3D<int>* voxels =
        dynamic_cast<NTensorField3D<int>*>(blocks[0]);
    PLB_ASSERT( voxels );
    AtomicContainerBlock3D* container =
        dynamic_cast<AtomicContainerBlock3D*>(blocks[1]);
    PLB_ASSERT( container );

    // Return if this block is already voxelized.
    if (voxels->getFlag()) {
        return;
    }

    Array<plint,2> xRange, yRange, zRange;
    if (!createVoxelizationRange(domain, *voxels, xRange, yRange, zRange)) {
        // If no seed has been found in the envelope, just return and wait
        //   for the next round.
        return;
    }

    // Specify if the loops go in positive or negative direction.
    plint xIncr = xRange[1]>xRange[0] ? 1 : -1;
    plint yIncr = yRange[1]>yRange[0] ? 1 : -1;
    plint zIncr = zRange[1]>zRange[0] ? 1 : -1;
    // The ranges are closed on both ends. Here, the range[1] end
    //   is converted to an open one so we can use != checks in the loops.
    xRange[1] += xIncr; 
    yRange[1] += yIncr; 
    zRange[1] += zIncr; 
    for (plint iX=xRange[0]; iX!=xRange[1]; iX+=xIncr) {
        for (plint iY=yRange[0]; iY!=yRange[1]; iY+=yIncr) {
            for (plint iZ=zRange[0]; iZ!=zRange[1]; iZ+=zIncr) {
                Dot3D pos(iX,iY,iZ);
                int voxelType = *voxels->get(iX,iY,iZ);
                if (voxelType==voxelFlag::undetermined) {
                    for (plint dx=-1; dx<=+1; ++dx) {
                        for (plint dy=-1; dy<=+1; ++dy) {
                            for (plint dz=-1; dz<=+1; ++dz) {
                                if (!(dx==0 && dy==0 && dz==0)) {
                                    Dot3D neighbor(iX+dx, iY+dy, iZ+dz);
                                    bool ok = voxelizeFromNeighbor (
                                                  *voxels, *container, 
                                                  pos, neighbor, voxelType );
                                    if (!ok) {
                                        printOffender(*voxels, *container, pos);
                                    }
                                    PLB_ASSERT( ok );
                                }
                            }
                        }
                    }
                    *voxels->get(iX,iY,iZ) = voxelType;
                }
            }
        }
    }
    // Indicate that this atomic-block has been voxelized.
    voxels->setFlag(true);
}

template<typename T>
MeshToVoxelFunctional3D<T>* MeshToVoxelFunctional3D<T>::clone() const {
    return new MeshToVoxelFunctional3D<T>(*this);
}

template<typename T>
void MeshToVoxelFunctional3D<T>::getTypeOfModification(std::vector<modif::ModifT>& modified) const {
    modified[0] = modif::staticVariables;  // Voxels
    modified[1] = modif::nothing; // Hash Container
}

template<typename T>
BlockDomain::DomainT MeshToVoxelFunctional3D<T>::appliesTo() const {
    return BlockDomain::bulk;
}

/* ***************** Robust Voxelizer ********************************************** */

template<typename T>
MultiNTensorField3D<int>* voxelizeMesh (
        ConnectedTriangleMesh<T>& mesh,
        plint symmetricLayer, plint borderWidth )
{
    Cuboid<T> cuboid = mesh.getBoundingCuboid();
    T dX = cuboid.upperRightCorner[0] - cuboid.lowerLeftCorner[0];
    T dY = cuboid.upperRightCorner[1] - cuboid.lowerLeftCorner[1];
    T dZ = cuboid.upperRightCorner[2] - cuboid.lowerLeftCorner[2];

    // Creation of the multi-scalar field. The +1 is because if the resolution is N,
    //   the number of nodes is N+1.
    plint nx = (plint) dX + 1 + 2*symmetricLayer;
    plint ny = (plint) dY + 1 + 2*symmetricLayer;
    plint nz = (plint) dZ + 1 + 2*symmetricLayer;

    return voxelizeMesh(mesh, Box3D(0,nx-1, 0,ny-1, 0,nz-1), borderWidth);
}

template<typename T>
MultiNTensorField3D<int>* voxelizeMesh (
        ConnectedTriangleMesh<T>& mesh,
        Box3D const& domain, plint borderWidth )
{
    // As initial seed, a one-cell layer around the outer boundary is tagged
    //   as outside cells.
    plint envelopeWidth = 1;
    plint nDim = 1;
    int iniVal = voxelFlag::outside;
    MultiNTensorField3D<int>* voxelMatrix
        = generateMultiNTensorField3D<int>(domain, nDim, &iniVal, envelopeWidth);
    setToConstant( *voxelMatrix, voxelMatrix->getBoundingBox().enlarge(-1),
                   voxelFlag::undetermined );

    MultiContainerBlock3D hashContainer(*voxelMatrix);
    std::vector<MultiBlock3D*> container_arg;
    container_arg.push_back(&hashContainer);
    applyProcessingFunctional (
            new MeshToHash<T>(mesh),
            hashContainer.getBoundingBox(), container_arg );

    std::vector<MultiBlock3D*> flag_hash_arg;
    flag_hash_arg.push_back(voxelMatrix);
    flag_hash_arg.push_back(&hashContainer);

    plint sumVoxel;
    computeIntSum(*voxelMatrix, voxelMatrix->getBoundingBox(), &sumVoxel, nDim);
    bool done = false;
    plint maxIter = 5000;
    plint i=0;
    while (!done && i<maxIter) {
        applyProcessingFunctional (
                new Voxelizer3D<T>(mesh),
                voxelMatrix->getBoundingBox(), flag_hash_arg );
        plint nextSumVoxel;
        computeIntSum(*voxelMatrix, voxelMatrix->getBoundingBox(), &nextSumVoxel, nDim);
        if (nextSumVoxel == sumVoxel) {
            done = true;
        }
        sumVoxel = nextSumVoxel;
        ++i;
    }
    if (i==maxIter) {
        pcout << "Warning: Voxelization failed." << std::endl;
    }

    setUndeterminedToFlag(*voxelMatrix, voxelFlag::inside);

    // Add the "innerBorder" and "outerBorder" flags.
    identifyBoundary(*voxelMatrix, voxelMatrix->getBoundingBox().enlarge(-1), borderWidth);

    return voxelMatrix;
}

template<typename T>
MultiNTensorField3D<int>* voxelizeMesh (
        ConnectedTriangleMesh<T>& mesh,
        Box3D const& domain, plint borderWidth, Box3D seed )
{
    // As initial seed, a one-cell layer around the outer boundary is tagged
    //   as ouside cells.
    plint envelopeWidth=1;
    plint nDim = 1;
    int iniVal = voxelFlag::undetermined;
    MultiNTensorField3D<int>* voxelMatrix
        = generateMultiNTensorField3D<int>(domain, nDim, &iniVal, envelopeWidth);
    setToConstant(*voxelMatrix, seed, voxelFlag::outside);

    MultiContainerBlock3D hashContainer(*voxelMatrix);
    std::vector<MultiBlock3D*> container_arg;
    container_arg.push_back(&hashContainer);
    applyProcessingFunctional (
            new MeshToHash<T>(mesh),
            hashContainer.getBoundingBox(), container_arg );

    std::vector<MultiBlock3D*> flag_hash_arg;
    flag_hash_arg.push_back(voxelMatrix);
    flag_hash_arg.push_back(&hashContainer);

    plint sumVoxel;
    computeIntSum(*voxelMatrix, voxelMatrix->getBoundingBox(), &sumVoxel, nDim);
    bool done = false;
    plint maxIter = 5000;
    plint i=0;
    while (!done && i<maxIter) {
        applyProcessingFunctional (
                new Voxelizer3D<T>(mesh),
                voxelMatrix->getBoundingBox(), flag_hash_arg );
        plint nextSumVoxel;
        computeIntSum(*voxelMatrix, voxelMatrix->getBoundingBox(), &nextSumVoxel, nDim);
        if (nextSumVoxel == sumVoxel) {
            done = true;
        }
        sumVoxel = nextSumVoxel;
        ++i;
    }
    if (i==maxIter) {
        pcout << "Warning: Voxelization failed." << std::endl;
    }

    setUndeterminedToFlag(*voxelMatrix, voxelFlag::inside);

    // Add the "innerBorder" and "outerBorder" flags.
    identifyBoundary(*voxelMatrix, voxelMatrix->getBoundingBox().enlarge(-1), borderWidth);

    return voxelMatrix;
}

/* ******** Voxelizer3D ************************************* */

template<typename T>
Voxelizer3D<T>::Voxelizer3D (
        ConnectedTriangleMesh<T>& mesh_)
    : mesh(mesh_)
{ }

template<typename T>
bool Voxelizer3D<T>::checkIfFacetsCrossed (
        AtomicContainerBlock3D& hashContainer,
        Array<T,3> const& point1, Array<T,3> const& point2,
        T& distance, plint& whichTriangle )
{
    Array<T,2> xRange (
                 std::min(point1[0], point2[0]),
                 std::max(point1[0], point2[0]) );
    Array<T,2> yRange (
                 std::min(point1[1], point2[1]),
                 std::max(point1[1], point2[1]) );
    Array<T,2> zRange (
                 std::min(point1[2], point2[2]),
                 std::max(point1[2], point2[2]) );
    TriangleHashing<T> triangleHash(hashContainer);
    std::vector<plint> possibleTriangles;
    triangleHash.getTriangles(xRange, yRange, zRange, possibleTriangles);

    int flag = 0; // Check for crossings inside the point1-point2 segment.
    Array<T,3> intersection; // Dummy variable.
    Array<T,3> normal;       // Dummy variable.
    T tmpDistance;           // Dummy variable.

    for (pluint iPossible=0; iPossible<possibleTriangles.size(); ++iPossible) {
        plint triangleID = possibleTriangles[iPossible];
        typename ConnectedTriangleMesh<T>::PTriangle triangle = mesh.triangle(triangleID);
        if (pointOnTriangle(point1, point2, flag, *triangle, intersection, normal, tmpDistance)==1) {
            distance = tmpDistance;
            whichTriangle = triangleID;
            return true;
        }
    }
    return false;
}

template<typename T>
bool Voxelizer3D<T>::createVoxelizationRange (
        Box3D const& domain, NTensorField3D<int>& voxels,
        Array<plint,2>& xRange, Array<plint,2>& yRange, Array<plint,2>& zRange )
{
    // The purpose of the first three loops is to locate the eight
    //   corners of the cube. One voxel per corner would be insufficient
    //   because a potential seed is situated differently, depending on
    //   whether it is on the boundary of the multi-block or somewhere inside.
    for (plint dx=0; dx<=+1; ++dx) {
        plint xMin = domain.x0+dx*domain.getNx()-1;
        plint xMax = domain.x0+dx*domain.getNx();
        for (plint dy=0; dy<=+1; ++dy) {
            plint yMin = domain.y0+dy*domain.getNy()-1;
            plint yMax = domain.y0+dy*domain.getNy();
            for (plint dz=0; dz<=+1; ++dz) {
                plint zMin = domain.z0+dz*domain.getNz()-1;
                plint zMax = domain.z0+dz*domain.getNz();

                // Locate a potential seed in one of the corners.
                for (plint iX=xMin; iX<=xMax; ++iX) {
                    for (plint iY=yMin; iY<=yMax; ++iY) {
                        for (plint iZ=zMin; iZ<=zMax; ++iZ) {
                            if (*voxels.get(iX,iY,iZ) != voxelFlag::undetermined) {
                                xRange[0] = domain.x0+dx*(domain.getNx()-1);
                                xRange[1] = domain.x0+(1-dx)*(domain.getNx()-1);
                                yRange[0] = domain.y0+dy*(domain.getNy()-1);
                                yRange[1] = domain.y0+(1-dy)*(domain.getNy()-1);
                                zRange[0] = domain.z0+dz*(domain.getNz()-1);
                                zRange[1] = domain.z0+(1-dz)*(domain.getNz()-1);
                                return true;
                            }
                        }
                    }
                }

            }
        }
    }
    return false;
}

template<typename T>
void Voxelizer3D<T>::printOffender (
        NTensorField3D<int> const& voxels,
        AtomicContainerBlock3D& hashContainer,
        Dot3D pos )
{
    std::set<plint> triangles;
    Dot3D offset = voxels.getLocation();
    Dot3D pos_ = pos+offset;
    std::cout << "Position (" << pos_.x << "," << pos_.y << "," << pos_.z << ")" << std::endl;
    for (plint dx=-1; dx<=+1; ++dx) {
        for (plint dy=-1; dy<=+1; ++dy) {
            for (plint dz=-1; dz<=+1; ++dz) {
                if (!(dx==0 && dy==0 && dz==0)) {
                    Dot3D neigh = pos+offset+Dot3D(dx,dy,dz);
                    int typeOfNeighbor = *voxels.get(pos.x+dx,pos.y+dy,pos.z+dz);
                    if (typeOfNeighbor!=voxelFlag::undetermined) {
                        T distance;
                        plint whichTriangle;
                        Array<T,3> p1(pos_.x,pos_.y,pos_.z);
                        Array<T,3> p2(neigh.x,neigh.y,neigh.z);
                        global::counter("voxelizer-debug").increment(1);
                        bool crossed = checkIfFacetsCrossed (
                                    hashContainer, p1, p2, distance, whichTriangle);
                        global::counter("voxelizer-debug").reset();
                        std::cout << "Neighbor ("
                                  << dx << "," << dy << "," << dz
                                  << "); is "
                                  << (voxelFlag::insideFlag(typeOfNeighbor) ? "inside" : "outside");
                        if (crossed) {
                            triangles.insert(whichTriangle);
                            std::cout 
                                  << " inters. at distance " << distance
                                  << " with triangle " << whichTriangle << std::endl;
                        }
                        else {
                            std::cout << " no inters." << std::endl;
                        }
                    }
                }
            }
        }
    }
    std::set<plint>::iterator it = triangles.begin();
    for (; it!=triangles.end(); ++it) {
        std::cout << "Triangle " << *it << " [" << std::flush;
        typename ConnectedTriangleMesh<T>::PTriangle triangle = mesh.triangle(*it);
        Array<T,3> p0 = (*triangle)[0];
        Array<T,3> p1 = (*triangle)[1];
        Array<T,3> p2 = (*triangle)[2];
        std::cout << p0[0] << " " << p1[0] << " " << p2[0] << " " << p0[0] << "], ["
                  << p0[1] << " " << p1[1] << " " << p2[1] << " " << p0[1] << "], ["
                  << p0[2] << " " << p1[2] << " " << p2[2] << " " << p0[2] << "]" << std::endl;
    }
}

template<typename T>
bool Voxelizer3D<T>::voxelizeFromNeighbor (
        NTensorField3D<int> const& voxels,
        AtomicContainerBlock3D& hashContainer,
        Dot3D pos, Dot3D neighbor, int& voxelType )
{
    int verificationLevel = 0;
    Dot3D offset = voxels.getLocation();
    int typeOfNeighbor = *voxels.get(neighbor.x,neighbor.y,neighbor.z);
    if (typeOfNeighbor==voxelFlag::undetermined) {
        return true;
    }
    // If there is no verification and the voxel has already been voxelized,
    //   it is not being re-voxelized here.
    if (verificationLevel==0) {
        if (voxelType!=voxelFlag::undetermined) {
            return true;
        }
    }
    Dot3D pos_ = pos+offset;
    Dot3D neighbor_ = neighbor+offset;
    Array<T,3> point1((T)pos_.x, (T)pos_.y, (T)pos_.z);
    Array<T,3> point2((T)neighbor_.x, (T)neighbor_.y, (T)neighbor_.z);
    int newVoxelType = voxelFlag::undetermined;
    T distance1;
    plint whichTriangle1;
    if (checkIfFacetsCrossed(hashContainer, point1, point2, distance1, whichTriangle1)) {
        return true;
    }
    else {
        newVoxelType = typeOfNeighbor;
    }
    int oldVoxelType = voxelType;
    voxelType = newVoxelType;
    if (oldVoxelType == voxelFlag::undetermined) {
        return true;
    }
    else {
        return oldVoxelType == newVoxelType;
    }
}

template<typename T>
void Voxelizer3D<T>::processGenericBlocks (
        Box3D domain, std::vector<AtomicBlock3D*> blocks )
{
    PLB_PRECONDITION( blocks.size()==2 );
    NTensorField3D<int>* voxels =
        dynamic_cast<NTensorField3D<int>*>(blocks[0]);
    PLB_ASSERT( voxels );
    AtomicContainerBlock3D* container =
        dynamic_cast<AtomicContainerBlock3D*>(blocks[1]);
    PLB_ASSERT( container );

    Array<plint,2> xRange, yRange, zRange;
    if (!createVoxelizationRange(domain, *voxels, xRange, yRange, zRange)) {
        // If no seed has been found in the envelope, just return and wait
        //   for the next round.
        return;
    }

    // Specify if the loops go in positive or negative direction.
    plint xIncr = xRange[1]>xRange[0] ? 1 : -1;
    plint yIncr = yRange[1]>yRange[0] ? 1 : -1;
    plint zIncr = zRange[1]>zRange[0] ? 1 : -1;
    // The ranges are closed on both ends. Here, the range[1] end
    //   is converted to an open one so we can use != checks in the loops.
    xRange[1] += xIncr; 
    yRange[1] += yIncr; 
    zRange[1] += zIncr; 
    for (plint iX=xRange[0]; iX!=xRange[1]; iX+=xIncr) {
        for (plint iY=yRange[0]; iY!=yRange[1]; iY+=yIncr) {
            for (plint iZ=zRange[0]; iZ!=zRange[1]; iZ+=zIncr) {
                Dot3D pos(iX,iY,iZ);
                int voxelType = *voxels->get(iX,iY,iZ);
                if (voxelType==voxelFlag::undetermined) {
                    for (plint dx=-1; dx<=+1; ++dx) {
                        for (plint dy=-1; dy<=+1; ++dy) {
                            for (plint dz=-1; dz<=+1; ++dz) {
                                if (!(dx==0 && dy==0 && dz==0)) {
                                    Dot3D neighbor(iX+dx, iY+dy, iZ+dz);
                                    bool ok = voxelizeFromNeighbor (
                                                  *voxels, *container, 
                                                  pos, neighbor, voxelType );
                                    if (!ok) {
                                        printOffender(*voxels, *container, pos);
                                    }
                                    PLB_ASSERT( ok );
                                    if (voxelType != voxelFlag::undetermined) {
                                        goto out;
                                    }
                                }
                            }
                        }
                    }
out:
                    *voxels->get(iX,iY,iZ) = voxelType;
                }
            }
        }
    }
}

template<typename T>
Voxelizer3D<T>* Voxelizer3D<T>::clone() const {
    return new Voxelizer3D<T>(*this);
}

template<typename T>
void Voxelizer3D<T>::getTypeOfModification(std::vector<modif::ModifT>& modified) const {
    modified[0] = modif::staticVariables;  // Voxels
    modified[1] = modif::nothing; // Hash Container
}

template<typename T>
BlockDomain::DomainT Voxelizer3D<T>::appliesTo() const {
    return BlockDomain::bulk;
}



/* ******** MakeSparse ************************************* */

inline MultiNTensorField3D<int>* makeSparse (
        MultiNTensorField3D<int>& fullVoxelMatrix,
        MultiNTensorField3D<int>& mask, int margin,
        plint blockSize, plint envelopeWidth )
{
    MultiNTensorField3D<int> domainMatrix(1, (MultiBlock3D const&) fullVoxelMatrix);
    maskedSetToConstant(domainMatrix, mask, domainMatrix.getBoundingBox(), 1);
    for (int iLayer=0; iLayer<margin; ++iLayer) {
        growDomain(domainMatrix, domainMatrix.getBoundingBox(), 1);
    }
    MultiNTensorField3D<int>* multiBlockDomainMatrix =
        plb::reparallelize(domainMatrix, blockSize,blockSize,blockSize);
    int eliminateFlag = 0;
    MultiBlockManagement3D sparseBlockManagement = 
        sparseManagement ( *multiBlockDomainMatrix, eliminateFlag, envelopeWidth );
    delete multiBlockDomainMatrix;
    int nDim = 1;
    int iniVal = (int)voxelFlag::undetermined;
    MultiNTensorField3D<int>* voxelMatrix = new MultiNTensorField3D<int> (
            nDim, &iniVal, 
            sparseBlockManagement,
            fullVoxelMatrix.getBlockCommunicator().clone(),
            fullVoxelMatrix.getCombinedStatistics().clone(),
            defaultMultiBlockPolicy3D().getMultiNTensorAccess<int>() );
    copy(fullVoxelMatrix, fullVoxelMatrix.getBoundingBox(), *voxelMatrix, voxelMatrix->getBoundingBox());
    return voxelMatrix;
}

/// Create a 1/0 mask for inside/innerborder resp.
/// outside/outerborder regions.
inline MultiNTensorField3D<int>* fluidMask (
        MultiNTensorField3D<int>& fullVoxelMatrix, int flowType )
{
    MultiNTensorField3D<int>* mask1 = equals (
            fullVoxelMatrix, &flowType, 1, fullVoxelMatrix.getBoundingBox() );
    int flowBoundary = voxelFlag::borderFlag(flowType);
    MultiNTensorField3D<int>* mask2 = equals (
            fullVoxelMatrix, &flowBoundary, 1, fullVoxelMatrix.getBoundingBox() );
    MultiNTensorField3D<int>* mask = logicalOr (
            *mask1, *mask2, fullVoxelMatrix.getBoundingBox() );
    delete mask1;
    delete mask2;
    return mask;
}

inline MultiNTensorField3D<int>* createSparseVoxelMatrix (
        MultiNTensorField3D<int>& fullVoxelMatrix, int flowType,
        int margin, plint blockSize, plint envelopeWidth)
{
    if (blockSize>0) {
        MultiNTensorField3D<int>* mask = fluidMask(fullVoxelMatrix, flowType);
        MultiNTensorField3D<int>* sparseVoxelMatrix =
            makeSparse(fullVoxelMatrix, *mask, margin, blockSize, envelopeWidth);
        delete mask;
        return sparseVoxelMatrix;
    }
    else {
        // blockSize=0 means: don't create a sparse representation of the
        // multi-block. The envelope of the voxel-matrix needs to be extended
        // for future usage, though.
        MultiNTensorField3D<int>* voxelMatrix = generateMultiNTensorField3D<int> (
                        (MultiBlock3D&) fullVoxelMatrix, envelopeWidth, fullVoxelMatrix.getNdim() );
        copy(fullVoxelMatrix, fullVoxelMatrix.getBoundingBox(), *voxelMatrix, voxelMatrix->getBoundingBox());
        return voxelMatrix;
    }
}


/* ******** IdentifyBoundaryFunctional3D ************************************* */

template<typename T>
void identifyBoundary( MultiNTensorField3D<T>& voxelMatrix,
                       Box3D const& domain, plint borderWidth )
{
    applyProcessingFunctional( new IdentifyBoundaryFunctional3D<T>(borderWidth),
                               domain, voxelMatrix );
}

template<typename T>
IdentifyBoundaryFunctional3D<T>::IdentifyBoundaryFunctional3D(plint borderWidth_)
    : borderWidth(borderWidth_)
{ }

template<typename T>
void IdentifyBoundaryFunctional3D<T>::process (
        Box3D domain, NTensorField3D<T>& voxels )
{
    std::vector<Array<plint,3> > directions;
    for (plint dx=-borderWidth; dx<=borderWidth; ++dx) {
        for (plint dy=-borderWidth; dy<=borderWidth; ++dy) {
            for (plint dz=-borderWidth; dz<=borderWidth; ++dz) {
                if(!(dx==0 && dy==0 && dz==0)) {
                    directions.push_back(Array<plint,3>(dx,dy,dz));
                }
            }
        }
    }
    for (plint iX = domain.x0; iX <= domain.x1; ++iX) {
        for (plint iY = domain.y0; iY <= domain.y1; ++iY) {
            for (plint iZ = domain.z0; iZ <= domain.z1; ++iZ) {
                int& voxel = *voxels.get(iX,iY,iZ);
                if (voxelFlag::insideFlag(voxel)) {
                    for (pluint i=0; i<directions.size(); ++i) {
                        Array<plint,3> const& direction = directions[i];
                        plint nextX = iX + direction[0];
                        plint nextY = iY + direction[1];
                        plint nextZ = iZ + direction[2];
                        if (voxelFlag::outsideFlag(*voxels.get(nextX,nextY,nextZ))) {
                            voxel = voxelFlag::innerBorder;
                            break;
                        }
                    }
                }
                if (voxelFlag::outsideFlag(voxel)) {
                    for (pluint i=0; i<directions.size(); ++i) {
                        Array<plint,3> const& direction = directions[i];
                        plint nextX = iX + direction[0];
                        plint nextY = iY + direction[1];
                        plint nextZ = iZ + direction[2];
                        if (voxelFlag::insideFlag(*voxels.get(nextX,nextY,nextZ))) {
                            voxel = voxelFlag::outerBorder;
                            break;
                        }
                    }
                }
            }
        }
    }
}

template<typename T>
IdentifyBoundaryFunctional3D<T>* IdentifyBoundaryFunctional3D<T>::clone() const {
    return new IdentifyBoundaryFunctional3D<T>(*this);
}

template<typename T>
void IdentifyBoundaryFunctional3D<T>::getTypeOfModification(std::vector<modif::ModifT>& modified) const {
    modified[0] = modif::staticVariables;
}

template<typename T>
BlockDomain::DomainT IdentifyBoundaryFunctional3D<T>::appliesTo() const {
    return BlockDomain::bulk;
}


/* ******** ReIdentifyBoundaryFunctional3D ************************************* */

template<typename T>
void reIdentifyBoundary( MultiNTensorField3D<T>& voxelMatrix,
                         MultiNTensorField3D<T>& oldVoxelMatrix,
                         Box3D const& domain, plint borderWidth )
{
    applyProcessingFunctional( new ReIdentifyBoundaryFunctional3D<T>(borderWidth),
                               domain, voxelMatrix, oldVoxelMatrix );
}

template<typename T>
ReIdentifyBoundaryFunctional3D<T>::ReIdentifyBoundaryFunctional3D(plint borderWidth_)
    : borderWidth(borderWidth_)
{ }

template<typename T>
void ReIdentifyBoundaryFunctional3D<T>::process (
        Box3D domain, NTensorField3D<T>& voxels, NTensorField3D<T>& oldVoxels )
{
    std::vector<Array<plint,3> > directions;
    for (plint dx=-borderWidth; dx<=borderWidth; ++dx) {
        for (plint dy=-borderWidth; dy<=borderWidth; ++dy) {
            for (plint dz=-borderWidth; dz<=borderWidth; ++dz) {
                if(!(dx==0 && dy==0 && dz==0)) {
                    directions.push_back(Array<plint,3>(dx,dy,dz));
                }
            }
        }
    }
    for (plint iX = domain.x0; iX <= domain.x1; ++iX) {
        for (plint iY = domain.y0; iY <= domain.y1; ++iY) {
            for (plint iZ = domain.z0; iZ <= domain.z1; ++iZ) {
                int oldVoxel = *oldVoxels.get(iX,iY,iZ);
                int& voxel = *voxels.get(iX,iY,iZ);
                if (oldVoxel==voxelFlag::undetermined && voxelFlag::insideFlag(voxel)) {
                    for (pluint i=0; i<directions.size(); ++i) {
                        Array<plint,3> const& direction = directions[i];
                        plint nextX = iX + direction[0];
                        plint nextY = iY + direction[1];
                        plint nextZ = iZ + direction[2];
                        if (voxelFlag::outsideFlag(*voxels.get(nextX,nextY,nextZ))) {
                            voxel = voxelFlag::innerBorder;
                            break;
                        }
                    }
                }
                if (voxelFlag::outsideFlag(voxel)) {
                    for (pluint i=0; i<directions.size(); ++i) {
                        Array<plint,3> const& direction = directions[i];
                        plint nextX = iX + direction[0];
                        plint nextY = iY + direction[1];
                        plint nextZ = iZ + direction[2];
                        if (voxelFlag::insideFlag(*voxels.get(nextX,nextY,nextZ))) {
                            voxel = voxelFlag::outerBorder;
                            break;
                        }
                    }
                }
            }
        }
    }
}

template<typename T>
ReIdentifyBoundaryFunctional3D<T>* ReIdentifyBoundaryFunctional3D<T>::clone() const {
    return new ReIdentifyBoundaryFunctional3D<T>(*this);
}

template<typename T>
void ReIdentifyBoundaryFunctional3D<T>::getTypeOfModification(std::vector<modif::ModifT>& modified) const {
    modified[0] = modif::staticVariables;
    modified[1] = modif::nothing;
}

template<typename T>
BlockDomain::DomainT ReIdentifyBoundaryFunctional3D<T>::appliesTo() const {
    return BlockDomain::bulk;
}


/* ******** GrowDomainFunctional3D ************************************* */

template<typename T>
void growDomain( MultiNTensorField3D<T>& matrix,
                 Box3D const& domain, T flag )
{
    applyProcessingFunctional( new GrowDomainFunctional3D<T>(flag),
                               domain, matrix );
}

template<typename T>
GrowDomainFunctional3D<T>::GrowDomainFunctional3D(T flag_)
    : flag(flag_)
{ }

template<typename T>
void GrowDomainFunctional3D<T>::process (
        Box3D domain, NTensorField3D<T>& voxels )
{
    PLB_ASSERT( voxels.getNdim()==1 );
    std::vector<Array<plint,3> > directions;
    for (plint dx=-1; dx<=1; ++dx) {
        for (plint dy=-1; dy<=1; ++dy) {
            for (plint dz=-1; dz<=1; ++dz) {
                if(!(dx==0 && dy==0 && dz==0)) {
                    directions.push_back(Array<plint,3>(dx,dy,dz));
                }
            }
        }
    }
    std::vector<Dot3D> boundaryCells;
    for (plint iX = domain.x0; iX <= domain.x1; ++iX) {
        for (plint iY = domain.y0; iY <= domain.y1; ++iY) {
            for (plint iZ = domain.z0; iZ <= domain.z1; ++iZ) {
                if (*voxels.get(iX,iY,iZ) != flag) {
                    for (pluint i=0; i<directions.size(); ++i) {
                        Array<plint,3> const& direction = directions[i];
                        plint nextX = iX + direction[0];
                        plint nextY = iY + direction[1];
                        plint nextZ = iZ + direction[2];
                        if (*voxels.get(nextX,nextY,nextZ)==flag) {
                            boundaryCells.push_back(Dot3D(iX,iY,iZ));
                            break;
                        }
                    }
                }
            }
        }
    }
    for (pluint i=0; i<boundaryCells.size(); ++i) {
        Dot3D cell = boundaryCells[i];
        *voxels.get(cell.x, cell.y, cell.z) = flag;
    }
}

template<typename T>
GrowDomainFunctional3D<T>* GrowDomainFunctional3D<T>::clone() const {
    return new GrowDomainFunctional3D<T>(*this);
}

template<typename T>
void GrowDomainFunctional3D<T>::getTypeOfModification(std::vector<modif::ModifT>& modified) const {
    modified[0] = modif::staticVariables;
}

template<typename T>
BlockDomain::DomainT GrowDomainFunctional3D<T>::appliesTo() const {
    return BlockDomain::bulk;
}

/* ******** SparseManagementFunctional3D ************************************ */

struct KeepBlockTag3D : public ContainerBlockData {
    bool keepThisBlock;
    virtual KeepBlockTag3D* clone() const {
        return new KeepBlockTag3D(*this);
    }
};

template<typename T>
SparseManagementFunctional3D<T>::SparseManagementFunctional3D(T eliminateFlag_)
    : eliminateFlag(eliminateFlag_),
      numBlocksId(this->getStatistics().subscribeIntSum())
{ }

template<typename T>
void SparseManagementFunctional3D<T>::processGenericBlocks (
        Box3D domain, std::vector<AtomicBlock3D*> blocks )
{
    PLB_PRECONDITION( blocks.size()==2 );
    NTensorField3D<T>* field = dynamic_cast<NTensorField3D<T>*>(blocks[0]);
    PLB_ASSERT( field );
    PLB_ASSERT( field->getNdim()==1 );
    AtomicContainerBlock3D* container = dynamic_cast<AtomicContainerBlock3D*>(blocks[1]);
    PLB_ASSERT( container );
    bool exclusivelyEliminateCells = true;
    for (plint iX=domain.x0; iX<=domain.x1; ++iX) {
        for (plint iY=domain.y0; iY<=domain.y1; ++iY) {
            for (plint iZ=domain.z0; iZ<=domain.z1; ++iZ) {
                if (*field->get(iX,iY,iZ) != eliminateFlag) {
                    exclusivelyEliminateCells = false;
                }
            }
        }
    }
    KeepBlockTag3D* flagData = new KeepBlockTag3D;
    if (exclusivelyEliminateCells) {
        flagData->keepThisBlock = false;
        this->getStatistics().gatherIntSum(numBlocksId, 1);
    }
    else {
        flagData->keepThisBlock = true;
    }
    container->setData(flagData);
}

template<typename T>
SparseManagementFunctional3D<T>* SparseManagementFunctional3D<T>::clone() const {
    return new SparseManagementFunctional3D<T>(*this);
}

template<typename T>
void SparseManagementFunctional3D<T>::getTypeOfModification(std::vector<modif::ModifT>& modified) const {
    modified[0] = modif::nothing; // Scalar Field.
    modified[1] = modif::staticVariables;  // Container Block with flag data.
}

template<typename T>
BlockDomain::DomainT SparseManagementFunctional3D<T>::appliesTo() const {
    return BlockDomain::bulk;
}

template<typename T>
pluint SparseManagementFunctional3D<T>::getNumBlocks() const {
    return this->getStatistics().getIntSum(numBlocksId);
}


/* ******** sparseManagement ************************************ */

template<typename T>
MultiBlockManagement3D sparseManagement (
        MultiNTensorField3D<T>& field, T eliminateFlag, plint newEnvelopeWidth )
{
    PLB_ASSERT( field.getNdim()==1 );
    MultiContainerBlock3D multiFlagBlock(field);
    std::vector<MultiBlock3D*> args;
    args.push_back(&field);
    args.push_back(&multiFlagBlock);
    SparseManagementFunctional3D<T> sparsityFunctional(eliminateFlag);
    applyProcessingFunctional (
            sparsityFunctional, field.getBoundingBox(), args );

    MultiBlockManagement3D const& management = multiFlagBlock.getMultiBlockManagement();
    ThreadAttribution const& threadAttribution = management.getThreadAttribution();
    SparseBlockStructure3D const& sparseBlock = management.getSparseBlockStructure();

    std::map<plint,Box3D> const& domains = sparseBlock.getBulks();
    std::vector<plint> domainIds(domains.size());
    std::vector<int> keepThisBlock(domains.size());

    std::map<plint,Box3D>::const_iterator it = domains.begin();
    plint pos = 0;
    for (; it != domains.end(); ++it) {
        plint id = it->first;
        domainIds[pos] = id;
        if (threadAttribution.isLocal(id)) {
            AtomicContainerBlock3D const& flagBlock = multiFlagBlock.getComponent(id);
            KeepBlockTag3D const* data =
                dynamic_cast<KeepBlockTag3D const*> (flagBlock.getData());
            PLB_ASSERT( data );
            keepThisBlock[pos] = data->keepThisBlock ? 1:0;
        }
        else {
            keepThisBlock[pos] = 0;
        }
        ++pos;
    }

#ifdef PLB_MPI_PARALLEL
    std::vector<int> tmp(keepThisBlock.size());
    global::mpi().reduceVect(keepThisBlock, tmp, MPI_SUM);
    global::mpi().bCast(&tmp[0], tmp.size());
    tmp.swap(keepThisBlock);
#endif

    SparseBlockStructure3D newSparseBlock(field.getBoundingBox());
    plint newId = 0;
    for (pluint iBlock=0; iBlock<keepThisBlock.size(); ++iBlock) {
        if (keepThisBlock[iBlock]) {
            plint id = domainIds[iBlock];
            Box3D bulk, uniqueBulk;
            sparseBlock.getBulk(id, bulk);
            sparseBlock.getUniqueBulk(id, uniqueBulk);
            newSparseBlock.addBlock(bulk, uniqueBulk, newId++);
        }
    }
    // If this assertion fails, that means that the domain covered
    // by the sparse block-structure is empty.
    PLB_ASSERT( newId>0 );

    ExplicitThreadAttribution* newAttribution = new ExplicitThreadAttribution;
    std::vector<std::pair<plint,plint> > ranges;
    plint numRanges = std::min(newId, (plint)global::mpi().getSize());
    util::linearRepartition(0, newId-1, numRanges, ranges);
    
    std::vector<plint> localBlocks;
    for (pluint iProc=0; iProc<ranges.size(); ++iProc) {
        for (plint blockId=ranges[iProc].first; blockId<=ranges[iProc].second; ++blockId) {
            newAttribution -> addBlock(blockId, iProc);
            if (iProc==(pluint)global::mpi().getRank()) {
                localBlocks.push_back(blockId);
            }
        }
    }

    MultiBlockManagement3D newManagement (
            newSparseBlock, newAttribution,
            newEnvelopeWidth,
            management.getRefinementLevel() );
    return newManagement;
}


/* ******** BoundaryToUndetFunctional ************************************* */

template<typename T>
void BoundaryToUndetFunctional<T>::process (
        Box3D domain, NTensorField3D<T>& voxels )
{
    PLB_ASSERT( voxels.getNdim()==1 );
    for (plint iX = domain.x0; iX <= domain.x1; ++iX) {
        for (plint iY = domain.y0; iY <= domain.y1; ++iY) {
            for (plint iZ = domain.z0; iZ <= domain.z1; ++iZ) {
                int& v = *voxels.get(iX,iY,iZ);
                if(v!=voxelFlag::inside && v!=voxelFlag::outside) {
                    v = voxelFlag::undetermined;
                }
            }
        }
    }
}

template<typename T>
BoundaryToUndetFunctional<T>* BoundaryToUndetFunctional<T>::clone() const {
    return new BoundaryToUndetFunctional<T>(*this);
}

template<typename T>
void BoundaryToUndetFunctional<T>::getTypeOfModification(std::vector<modif::ModifT>& modified) const {
    modified[0] = modif::staticVariables;
}

template<typename T>
BlockDomain::DomainT BoundaryToUndetFunctional<T>::appliesTo() const {
    return BlockDomain::bulk;
}

} // namespace plb

#endif  // MESH_TO_VOXEL_HH

