/* This file is part of the Palabos library.
 *
 * Copyright (C) 2011-2013 FlowKit Sarl
 * Route d'Oron 2
 * 1010 Lausanne, Switzerland
 * E-mail contact: contact@flowkit.com
 *
 * The most recent release of Palabos can be downloaded at 
 * <http://www.palabos.org/>
 *
 * The library Palabos is free software: you can redistribute it and/or
 * modify it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * The library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef NTENSOR_POST_PROCESSING_3D_HH
#define NTENSOR_POST_PROCESSING_3D_HH

#include "core/globalDefs.h"
#include "boundary/nTensorPostProcessing3D.h"
#include <algorithm>

namespace plb {

template<typename T>
OffLatticePostProcessing<T>::OffLatticePostProcessing(Group3D& fields_)
    : fields(fields_)
{ }

template<typename T>
T OffLatticePostProcessing<T>::computeAverage(MultiNTensorField3D<T>& data)
{
    PLB_ASSERT( data.getNdim()==1 );
    T result;
    MultiNTensorField3D<int>& exclusionMatrix = fields.getNTensor<int>("solidFlags");
    MultiNTensorField3D<int>* inclusionMatrix = plb::negate(exclusionMatrix, exclusionMatrix.getBoundingBox());
    maskedComputeAverage(data, *inclusionMatrix, data.getBoundingBox(), &result, 1);
    delete inclusionMatrix;
    return result;
}

template<typename T>
Array<T,3> OffLatticePostProcessing<T>::computeAverageVect(MultiNTensorField3D<T>& data)
{
    PLB_ASSERT( data.getNdim()==3 );
    Array<T,3> result;
    MultiNTensorField3D<int>& exclusionMatrix = fields.getNTensor<T>("solidFlags");
    MultiNTensorField3D<int>* inclusionMatrix = plb::negate(exclusionMatrix, exclusionMatrix.getBoundingBox());
    maskedComputeAverage(data, *inclusionMatrix, data.getBoundingBox(), &result[0], 3);
    delete inclusionMatrix;
    return result;
}

template<typename T>
void OffLatticePostProcessing<T>::process(MultiNTensorField3D<T>& data)
{
    int ndim = (int)data.getNdim();
    std::vector<T> zero(ndim, T());
    MultiNTensorField3D<int>& exclusionMatrix = fields.getNTensor<int>("solidFlags");
    maskedSetToConstant(data, exclusionMatrix, exclusionMatrix.getBoundingBox(), &zero[0], ndim );
}

template<typename T, template<typename U> class Descriptor>
void offLatticeMacroscopic(Group3D& macroFields, Group3D& fields, Units3D& units, T rho0, T rho0LB)
{
    MultiBlockLattice3D<T,Descriptor>& lattice = fields.getLattice<T,Descriptor>("fluid");

    MultiNTensorField3D<T>* velocityNorm =
        pypal_computeVelocityNorm(lattice, lattice.getBoundingBox());
    MultiNTensorField3D<T>* velocity=
        pypal_computeVelocity(lattice, lattice.getBoundingBox());
    MultiNTensorField3D<T>* vorticity=
        computeVorticity(*velocity, lattice.getBoundingBox());

    T velConst = units.dx()/units.dt();
    Array<T,3> velConst3(velConst, velConst, velConst);
    T vortConst = (T)1. / units.dt();
    Array<T,3> vortConst3(vortConst, vortConst, vortConst);

    multiplyInPlace(*velocityNorm, &velConst, 1, velocityNorm->getBoundingBox());
    multiplyInPlace(*velocity, &velConst3[0], 3, velocity->getBoundingBox());
    multiplyInPlace(*vorticity, &vortConst3[0], 3, vorticity->getBoundingBox());

    MultiNTensorField3D<T>* pressure=
        pypal_computeDensity(lattice, lattice.getBoundingBox());
    
    T cs2 = Descriptor<T>::cs2;
    T presConst = rho0*cs2 * util::sqr(units.dx()/units.dt());
    subtractInPlace(*pressure, &rho0LB, 1, pressure->getBoundingBox());
    multiplyInPlace(*pressure, &presConst, 1, pressure->getBoundingBox());

    OffLatticePostProcessing<T> postProcessing(fields);
    postProcessing.process(*velocityNorm);
    postProcessing.process(*velocity);
    postProcessing.process(*pressure);
    postProcessing.process(*vorticity);

    macroFields.add(velocityNorm, "velocityNorm");
    macroFields.add(velocity, "velocity");
    macroFields.add(pressure, "pressure");
    macroFields.add(vorticity, "vorticity");
}

template<typename T, template<typename U> class Descriptor>
void offLatticeForce(Group3D& macroFields, Group3D& fields, Units3D& units, bool isCompressible, T rho0, T rho0LB)
{
    MultiBlockLattice3D<T,Descriptor>& lattice = fields.getLattice<T,Descriptor>("fluid");

    MultiNTensorField3D<T>* stress =
        pypal_computeStress(lattice, lattice.getBoundingBox(), rho0LB, isCompressible);

    T forceConvert = units.physForce(1., rho0);
    Array<T,6> forceConvert6;
    std::fill(&forceConvert6[0], &forceConvert6[0]+6, forceConvert);
    multiplyInPlace(*stress, &forceConvert6[0], 6, stress->getBoundingBox());

    OffLatticePostProcessing<T> postProcessing(fields);
    postProcessing.process(*stress);

    macroFields.add(stress, "stress");
}

template<typename T, template<typename U> class Descriptor>
void writeVTK( Group3D& fields, Units3D& units, int flowType, plint iT,
               plint padding, T rho0, T rho0LB )
{

    std::string fname(createFileName("volumeData", iT, padding));
    VtkImageOutput3D<T> vtkOut( fname, units.dx(), units.physDomain().lowerLeftCorner );
    Group3D macroFields;
    offLatticeMacroscopic<T,Descriptor>(macroFields, fields, units, rho0, rho0LB);
    vtkOut.template writeData<float>( macroFields.getNTensor<T>("velocityNorm"), "velocityNorm" );
    vtkOut.template writeData<float>( macroFields.getNTensor<T>("velocity"), "velocity" );
    vtkOut.template writeData<float>( macroFields.getNTensor<T>("pressure"), "pressure" );
    MultiNTensorField3D<int>& voxels = fields.getNTensor<int>("voxels");
    MultiNTensorField3D<T>* cVoxels = copyConvert<int,T>(voxels, voxels.getBoundingBox());
    vtkOut.template writeData<float>( *cVoxels, "voxels" );
    delete cVoxels;
}

}  // namespace plb

#endif  // NTENSOR_POST_PROCESSING_3D_HH

