/* This file is part of the Palabos library.
 *
 * Copyright (C) 2011-2015 FlowKit Sarl
 * Route d'Oron 2
 * 1010 Lausanne, Switzerland
 * E-mail contact: contact@flowkit.com
 *
 * The most recent release of Palabos can be downloaded at 
 * <http://www.palabos.org/>
 *
 * The library Palabos is free software: you can redistribute it and/or
 * modify it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * The library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef OFF_LATTICE_SETUP_3D_HH
#define OFF_LATTICE_SETUP_3D_HH

#include "core/globalDefs.h"
#include "boundary/offLatticeSetup3D.h"

namespace plb {

// borderWidth: This is for the voxelizer: thickness of the layers which are tagged
//   "innerBorder" and "outerBorder". A thickness of 1 is OK for Guo, which needs
//   just one layer of cells.
// blockSize: Size of blocks in the sparse/parallel representation.
//     Zero means: don't use sparse representation.
//     Negative means: default to 16 for exterior flows and 0 to interior flows.
template<typename T>
void voxelizeDomain( Group3D& group, Units3D units, RawConnectedTriangleMesh<T>& mesh,
                     int flowType, plint borderWidth, plint blockSize, bool largeEnvelope )
{
    // For the computation of the sparse block-structure, we need to know the
    // "margin": an extra margin of allocated cells around the obstacle.
    // Requirement: margin>=borderWidth. For inner flows, it's nice to add some margin
    // to improve the Paraview visualization.
    plint margin = borderWidth;
    if (flowType==voxelFlag::inside) {
        margin += 2;
    }
    if (blockSize<0) {
        blockSize = (flowType==voxelFlag::inside) ? 16 : 0;
    }
    plint envelopeWidth = largeEnvelope ?
                                2 :  // Extrapolated off-lattice BCs.
                                1;

    // Voxelize the domain means: decide which lattice nodes are inside the obstacle and
    // which are outside.
    MultiNTensorField3D<int>* fullVoxelMatrix
        = meshToVoxel(mesh, units.lbDomain(), borderWidth);

    MultiNTensorField3D<int>* sparseVoxelMatrix = createSparseVoxelMatrix (
            *fullVoxelMatrix, flowType, margin, blockSize, envelopeWidth );
    delete fullVoxelMatrix; fullVoxelMatrix=0;
    group.add(sparseVoxelMatrix, "voxels");
    int solidFlag = voxelFlag::invert(flowType);
    MultiNTensorField3D<int>* solidFlagMatrix =
        equals(*sparseVoxelMatrix, &solidFlag, 1, sparseVoxelMatrix->getBoundingBox());
    group.add(solidFlagMatrix, "solidFlags");
}

template<typename T, template<typename U> class Descriptor>
void createFluidFields(Group3D& fields, Dynamics<T,Descriptor>* dynamics) {
    plint envelopeWidth1 = 1;
    plint envelopeWidth2 = 2;
    fields.generateLattice<T,Descriptor>("fluid", envelopeWidth2);
    fields.get("fluid").toggleInternalStatistics(false);
    defineDynamics(fields.getLattice<T,Descriptor>("fluid"), fields.getBoundingBox(), dynamics);
    maskedDefineDynamics( fields.getLattice<T,Descriptor>("fluid"),
                          fields.getNTensor<int>("solidFlags"),
                          fields.getBoundingBox(), new NoDynamics<T,Descriptor>() );

    // The rhoBar and j fields are used at both the collision and at the implementation of the
    // outflow boundary condition.
    fields.generateNTensor<T>("rhoBar", envelopeWidth1);
    fields.generateNTensor<T>("j", 3, envelopeWidth1);
    fields.get("rhoBar").toggleInternalStatistics(false);
    fields.get("j").toggleInternalStatistics(false);
}

template<typename T, class SurfaceData>
void createOfflatticeData( Group3D& fields, ConnectedTriangleMesh<T>& obstacleMesh,
                           OffLatticeModel3D<T,SurfaceData> const& offLatticeModel )
{
    fields.generateContainer("triangleHash");
    setupHashing(fields.getContainer("triangleHash"), obstacleMesh);

    // It is very important that the "offLatticeData" container block
    // has the same multi-block management as the lattice used in the
    // simulation, and also the same envelope-width.
    plint containerEnvWidth = fields.get("fluid").getMultiBlockManagement().getEnvelopeWidth();
    fields.generateContainer("offLatticeData", containerEnvWidth);
    Actions3D iniActions;
    iniActions.addBlock(fields.get("offLatticeData"));
    iniActions.addBlock(fields.get("voxels"));
    iniActions.addBlock(fields.get("triangleHash"));
    iniActions.addBlock(fields.get("fluid")); // As boundaryShapeArg.
    iniActions.addProcessor (
            new OffLatticePatternFunctional3D<T,SurfaceData> (
                offLatticeModel.clone() ),
            0, 1, 2, 3,
            fields.getBoundingBox() );
    // No communication. Just creates the container.

    iniActions.execute();
}

} // namespace plb

#endif  // OFF_LATTICE_SETUP_3D_HH

