/* This file is part of the Palabos library.
 *
 * Copyright (C) 2011-2013 FlowKit Sarl
 * Route d'Oron 2
 * 1010 Lausanne, Switzerland
 * E-mail contact: contact@flowkit.com
 *
 * The most recent release of Palabos can be downloaded at 
 * <http://www.palabos.org/>
 *
 * The library Palabos is free software: you can redistribute it and/or
 * modify it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * The library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef MESHES_AND_PARTICLES_H
#define MESHES_AND_PARTICLES_H

#include "core/units.h"
#include "particles/multiParticleField3D.h"
#include "latticeBoltzmann/nearestNeighborLattices3D.h"

namespace plb {

template<typename T>
void scalarFieldToMesh( MultiScalarField3D<T>& scalarField,
                        ConnectedTriangleMesh<T>& mesh, std::string propertyName );

template<typename T>
void nTensorFieldToMesh( MultiNTensorField3D<T>& nTensorField,
                         ConnectedTriangleMesh<T>& mesh, std::string propertyName,
                         MultiParticleField3D<DenseParticleField3D<T,descriptors::D3Q19Descriptor> >* particleField=0 );

template<typename T>
void nTensorFieldToMesh( MultiNTensorField3D<T>& nTensorField,
                         MultiNTensorField3D<int>& mask,
                         ConnectedTriangleMesh<T>& mesh, std::string propertyName,
                         MultiParticleField3D<DenseParticleField3D<T,descriptors::D3Q19Descriptor> >* particleField=0 );

template<typename T>
void stressToSurfaceForce( MultiNTensorField3D<T>& stress,
                           ConnectedTriangleMesh<T>& mesh, std::string propertyName,
                           MultiParticleField3D<DenseParticleField3D<T,descriptors::D3Q19Descriptor> >* particleField=0 );


template<typename T>
void stressToSurfaceForce( MultiNTensorField3D<T>& stress,
                           MultiNTensorField3D<int>& mask,
                           ConnectedTriangleMesh<T>& mesh, std::string propertyName,
                           MultiParticleField3D<DenseParticleField3D<T,descriptors::D3Q19Descriptor> >* particleField=0 );

template<typename T>
Array<T,3> surfaceForceIntegral( MultiNTensorField3D<T>& stress,
                                 MultiNTensorField3D<int>& mask,
                                 ConnectedTriangleMesh<T>& mesh,
                                 MultiParticleField3D<DenseParticleField3D<T,descriptors::D3Q19Descriptor> >* particleField=0 );

template<typename T>
Array<T,3> surfaceTorqueIntegral( MultiNTensorField3D<T>& stress,
                                  MultiNTensorField3D<int>& mask,
                                  ConnectedTriangleMesh<T>& mesh,
                                  Array<T,3> const& center,
                                  Array<T,3> const& unitaryAxis,
                                  Units3D const& units,
                                  MultiParticleField3D<DenseParticleField3D<T,descriptors::D3Q19Descriptor> >* particleField=0 );

template<typename T, template<typename U> class Descriptor, class ParticleFieldT>
void verticesToParticles( ConnectedTriangleMesh<T>& mesh,
                          MultiParticleField3D<ParticleFieldT>& particleField,
                          Particle3D<T,Descriptor>* particleTemplate, Box3D domain, bool copyVertexProperties=false );

template<typename T, template<typename U> class Descriptor, class ParticleFieldT>
void particlesToVertices( ConnectedTriangleMesh<T>& mesh,
                          MultiParticleField3D<ParticleFieldT>& particleField,
                          std::string propertyName, Box3D domain, plint nProperties=1 );

template<typename T, template<typename U> class Descriptor, class ParticleFieldT>
std::vector<T> computeSurfaceIntegral( MultiParticleField3D<ParticleFieldT>& particleField, ConnectedTriangleMesh<T>& mesh,
                                       plint nDim, Box3D domain );

template< typename T, template<typename U> class Descriptor >
class ParticleSurfaceIntegral3D : public PlainReductiveBoxProcessingFunctional3D
{
public:
    ParticleSurfaceIntegral3D(ConnectedTriangleMesh<T>& mesh_, plint nDim_);
    /// Arguments: [0] Particle-field
    virtual void processGenericBlocks(Box3D domain, std::vector<AtomicBlock3D*> fields);
    virtual ParticleSurfaceIntegral3D<T,Descriptor>* clone() const;
    virtual void getTypeOfModification(std::vector<modif::ModifT>& modified) const;
    std::vector<T> getIntegral() const;
private:
    ConnectedTriangleMesh<T>* mesh;
    plint nDim;
    std::vector<plint> integral_ids;
};

template<typename T, template<typename U> class Descriptor, class ParticleFieldT>
void particleStressToForce( MultiParticleField3D<ParticleFieldT>& particleField, ConnectedTriangleMesh<T>& mesh, Box3D domain);

template< typename T, template<typename U> class Descriptor >
class ParticleStressToForce3D : public BoxProcessingFunctional3D
{
public:
    ParticleStressToForce3D(ConnectedTriangleMesh<T>& mesh_);
    /// Arguments: [0] Particle-field
    virtual void processGenericBlocks(Box3D domain, std::vector<AtomicBlock3D*> fields);
    virtual ParticleStressToForce3D<T,Descriptor>* clone() const;
    virtual void getTypeOfModification(std::vector<modif::ModifT>& modified) const;
private:
    ConnectedTriangleMesh<T>* mesh;
};

template<typename T, template<typename U> class Descriptor, class ParticleFieldT>
void particleStressToTorque( MultiParticleField3D<ParticleFieldT>& particleField, ConnectedTriangleMesh<T>& mesh, Box3D domain,
        Array<T,3> const& center, Array<T,3> const& unitaryAxis);

template< typename T, template<typename U> class Descriptor >
class ParticleStressToTorque3D : public BoxProcessingFunctional3D
{
public:
    ParticleStressToTorque3D(ConnectedTriangleMesh<T>& mesh_, Array<T,3> const& center_, Array<T,3> const& unitaryAxis_);
    /// Arguments: [0] Particle-field
    virtual void processGenericBlocks(Box3D domain, std::vector<AtomicBlock3D*> fields);
    virtual ParticleStressToTorque3D<T,Descriptor>* clone() const;
    virtual void getTypeOfModification(std::vector<modif::ModifT>& modified) const;
private:
    ConnectedTriangleMesh<T>* mesh;
    Array<T,3> center, unitaryAxis;
};

template<typename T, template<typename U> class Descriptor, class ParticleFieldT>
void nTensorFieldToParticles(MultiNTensorField3D<T>& nTensor, MultiParticleField3D<ParticleFieldT>& particleField, Box3D domain);

template<typename T, template<typename U> class Descriptor, class ParticleFieldT>
void nTensorFieldToParticles( MultiNTensorField3D<T>& nTensor, MultiNTensorField3D<int>& mask,
                              MultiParticleField3D<ParticleFieldT>& particleField, Box3D domain);

template< typename T, template<typename U> class Descriptor >
class NTensorFieldToParticle3D : public BoxProcessingFunctional3D
{
public:
    NTensorFieldToParticle3D(plint nDim_, plint whichScalar_=-1);
    /// Arguments: [0] Particle-field; [1] NTensor-Field
    virtual void processGenericBlocks(Box3D domain, std::vector<AtomicBlock3D*> fields);
    virtual NTensorFieldToParticle3D<T,Descriptor>* clone() const;
    virtual void getTypeOfModification(std::vector<modif::ModifT>& modified) const;
private:
    plint nDim;
    plint whichScalar;
};

template< typename T, template<typename U> class Descriptor >
class MaskedNTensorFieldToParticle3D : public BoxProcessingFunctional3D
{
public:
    MaskedNTensorFieldToParticle3D(plint nDim_=1, plint whichScalar_=-1);
    /// Arguments: [0] Particle-field; [1] NTensor-Field; [2] Mask
    virtual void processGenericBlocks(Box3D domain, std::vector<AtomicBlock3D*> fields);
    virtual MaskedNTensorFieldToParticle3D<T,Descriptor>* clone() const;
    virtual void getTypeOfModification(std::vector<modif::ModifT>& modified) const;
private:
    plint nDim;
    plint whichScalar;
};

}  // namespace plb

#endif  // MESHES_AND_PARTICLES_H

