/* This file is part of the Palabos library.
 *
 * Copyright (C) 2011-2013 FlowKit Sarl
 * Route d'Oron 2
 * 1010 Lausanne, Switzerland
 * E-mail contact: contact@flowkit.com
 *
 * The most recent release of Palabos can be downloaded at 
 * <http://www.palabos.org/>
 *
 * The library Palabos is free software: you can redistribute it and/or
 * modify it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * The library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef PARTICLE_X3D_3D_HH
#define PARTICLE_X3D_3D_HH

#include "core/globalDefs.h"
#include "io/plbFiles.h"
#include "geometry/particleX3D.h"
#include "multiBlock/defaultMultiBlockPolicy3D.h"

#include <cstdio>
#include <cstdlib>
#include <sstream>
#include <iomanip>
#include <limits>

namespace plb {

template<typename T, template<typename U> class Descriptor>
void writeSurfaceX3D( TriangleBoundary3D<T> const& boundary,
                      MultiParticleField3D<DenseParticleField3D<T,Descriptor> >& particles,
                      std::vector<std::string> const& scalars,
                      std::string const& fName, bool dynamicMesh, plint tag )
{
    if (boundary.getMesh().getNumTriangles() == 0) {
        return;
    }
    writeSurfaceX3D(boundary, particles, scalars, fName, dynamicMesh, tag, std::vector<T>());
}

template<typename T, template<typename U> class Descriptor>
void writeSurfaceX3D( TriangleBoundary3D<T> const& boundary,
                      MultiParticleField3D<DenseParticleField3D<T,Descriptor> >& particles,
                      std::vector<std::string> const& scalars,
                      std::string const& fName, bool dynamicMesh, plint tag,
                      std::vector<T> const& scalarFactor )
{
    if (boundary.getMesh().getNumTriangles() == 0) {
        return;
    }
    SparseBlockStructure3D blockStructure(particles.getBoundingBox());
    blockStructure.addBlock(particles.getBoundingBox(), 0);
    plint envelopeWidth=1;
    MultiBlockManagement3D serialMultiBlockManagement (
            blockStructure, new OneToOneThreadAttribution, envelopeWidth );

    MultiParticleField3D<DenseParticleField3D<T,Descriptor> > multiSerialParticles (
            serialMultiBlockManagement,
            defaultMultiBlockPolicy3D().getCombinedStatistics() );

    copy ( particles, particles.getBoundingBox(), multiSerialParticles, particles.getBoundingBox() );
    if (global::mpi().isMainProcessor()) {
        ParticleField3D<T,Descriptor>& atomicSerialParticles =
            dynamic_cast<ParticleField3D<T,Descriptor>&>(multiSerialParticles.getComponent(0));

        std::vector<Particle3D<T,Descriptor>*> found;
        SmartBulk3D oneBlockBulk(serialMultiBlockManagement, 0);
        atomicSerialParticles.findParticles(oneBlockBulk.toLocal(particles.getBoundingBox()), found);

        x3dForVertices(found, boundary, scalars, fName, dynamicMesh, tag, scalarFactor);
    }
}

template<typename T, template<typename U> class Descriptor>
void exportX3D( TriangleBoundary3D<T> const& boundary,
                MultiParticleField3D<DenseParticleField3D<T,Descriptor> >& particles,
                std::vector<std::string> const& scalars, bool dynamicMesh, plint tag, FileName fileName )
{
    if (boundary.getMesh().getNumTriangles() == 0) {
        return;
    }
    exportX3D(boundary, particles, scalars, dynamicMesh, tag, std::vector<T>(), fileName);
}

template<typename T, template<typename U> class Descriptor>
void exportX3D( TriangleBoundary3D<T> const& boundary,
                MultiParticleField3D<DenseParticleField3D<T,Descriptor> >& particles,
                std::vector<std::string> const& scalars,
                bool dynamicMesh, plint tag, std::vector<T> const& scalarFactor, FileName fileName )
{
    if (boundary.getMesh().getNumTriangles() == 0) {
        return;
    }
    SparseBlockStructure3D blockStructure(particles.getBoundingBox());
    blockStructure.addBlock(particles.getBoundingBox(), 0);
    plint envelopeWidth=1;
    MultiBlockManagement3D serialMultiBlockManagement (
            blockStructure, new OneToOneThreadAttribution, envelopeWidth );

    MultiParticleField3D<DenseParticleField3D<T,Descriptor> > multiSerialParticles (
            serialMultiBlockManagement,
            defaultMultiBlockPolicy3D().getCombinedStatistics() );

    copy ( particles, particles.getBoundingBox(), multiSerialParticles, particles.getBoundingBox() );
    if (global::mpi().isMainProcessor()) {
        ParticleField3D<T,Descriptor>& atomicSerialParticles =
            dynamic_cast<ParticleField3D<T,Descriptor>&>(multiSerialParticles.getComponent(0));

        std::vector<Particle3D<T,Descriptor>*> found;
        SmartBulk3D oneBlockBulk(serialMultiBlockManagement, 0);
        atomicSerialParticles.findParticles(oneBlockBulk.toLocal(particles.getBoundingBox()), found);

        exportX3dForVertices(found, boundary, scalars, dynamicMesh, tag, scalarFactor, fileName);
    }
}

template<typename T, template<typename U> class Descriptor>
void x3dForVertices(std::vector<Particle3D<T,Descriptor>*> const& particles,
                    TriangleBoundary3D<T> const& boundary, std::vector<std::string> const& scalars,
                    std::string fName, bool dynamicMesh, plint tag, std::vector<T> const& scalarFactor)
{
    if (boundary.getMesh().getNumTriangles() == 0) {
        return;
    }
    PLB_ASSERT( scalarFactor.empty() || scalarFactor.size()==scalars.size() );
    if (dynamicMesh) {
        boundary.pushSelect(0,1); // 0=Open, 1=Dynamic.
    }
    else {
        boundary.pushSelect(0,0); // Open, Static.
    }
    TriangularSurfaceMesh<T> const& mesh = boundary.getMesh();
    // If this assertion fails, a likely explanation is that the margin of your sparse block
    // structure is too small, and one of the particles was outside the allocated domain.
    // PLB_PRECONDITION((plint)particles.size() == mesh.getNumVertices());
    if((plint)particles.size() != mesh.getNumVertices()) {
        pcout << "Warning: in x3dForVertices, the numer of particles doesn't match the number" << std::endl;
        pcout << "of vertices. There might be black spots in the produced X3D file." << std::endl;
    }

    std::vector<std::vector<T> > scalarData(scalars.size());
    for (pluint iScalar=0; iScalar<scalars.size(); ++iScalar) {
        scalarData[iScalar].resize(mesh.getNumVertices());
    }
    std::vector<Array<T,3> > posVect(mesh.getNumVertices());
    // Default initialize all data, just in case every vertex hasn't
    // an associated particle.
    for (plint i=0; i<mesh.getNumVertices(); ++i) {
        for (pluint iScalar=0; iScalar<scalars.size(); ++iScalar) {
            scalarData[iScalar][i] = T();
        }
        posVect[i].resetToZero();
    }
    T maxLimit = std::numeric_limits<T>::max();
    T minLimit = std::numeric_limits<T>::min();
    Array<T,3> minPos(maxLimit,maxLimit,maxLimit), maxPos(minLimit,minLimit,minLimit);
    for (pluint iParticle=0; iParticle<particles.size(); ++iParticle) {
        plint iVertex = particles[iParticle]->getTag();
        posVect[iVertex] = particles[iParticle]->getPosition();
        if (posVect[iVertex][0]<minPos[0]) minPos[0] = posVect[iVertex][0];
        if (posVect[iVertex][1]<minPos[1]) minPos[1] = posVect[iVertex][1];
        if (posVect[iVertex][2]<minPos[2]) minPos[2] = posVect[iVertex][2];
        if (posVect[iVertex][0]>maxPos[0]) maxPos[0] = posVect[iVertex][0];
        if (posVect[iVertex][1]>maxPos[1]) maxPos[1] = posVect[iVertex][1];
        if (posVect[iVertex][2]>maxPos[2]) maxPos[2] = posVect[iVertex][2];
        for (pluint iScalar=0; iScalar<scalars.size(); ++iScalar) {
            T scalar;
            particles[iParticle]->getScalar(iScalar, scalar);
            if (!scalarFactor.empty()) {
                scalar *= scalarFactor[iScalar];
            }
            scalarData[iScalar][iVertex] = scalar;
        }
    }
    T diffx = maxPos[0]-minPos[0];
    T diffy = maxPos[1]-minPos[1];
    T diffz = maxPos[2]-minPos[2];
    T diff = std::max(diffx,std::max(diffy,diffz));
    T dx_nonDim = (T)2./diff;
    Array<T,3> offset_nonDim (
                 -((T)1.0+minPos[0]/diff),
                 -((T)1.0+minPos[1]/diff),
                 -((T)1.0+minPos[2]/diff) );
    for (pluint iParticle=0; iParticle<particles.size(); ++iParticle) {
        plint iVertex = particles[iParticle]->getTag();
        posVect[iVertex] *= dx_nonDim;
        posVect[iVertex] += offset_nonDim;
    }
    T dx_ndToPhys = boundary.getDx()/dx_nonDim;
    Array<T,3> location_ndToPhys = boundary.getPhysicalLocation()-dx_ndToPhys*offset_nonDim;
    
    std::ofstream ofile(fName.c_str());
    ofile << std::fixed << std::setprecision(5);
    ofile << "<!DOCTYPE HTML>\n";
    ofile << "<html>\n";
    ofile << "    <head>\n";
    ofile << "        <title>" << "Palabos Geometry" << "</title>\n";
    ofile << "        <link rel=\"stylesheet\" type=\"text/css\" href=\"http://www.x3dom.org/download/x3dom.css\">\n";
    ofile << "        <script type=\"text/javascript\" src=\"http://www.x3dom.org/download/x3dom.js\">\n";
    ofile << "        </script>\n";

    ofile << "        <script type=\"text/javascript\">\n";
    ofile << "            var scale = " << std::setprecision(5) << dx_ndToPhys << ";\n";
    ofile << "            var offset = new Array(" << std::setprecision(5)
          << location_ndToPhys[0] << ","
          << location_ndToPhys[1] << ","
          << location_ndToPhys[2] << ");\n";
    ofile << "            var coloredProperties = new Object();\n";

    for (pluint iScalar=0; iScalar<scalars.size(); ++iScalar) {
        T minValue = std::numeric_limits<T>::max();
        T maxValue = std::numeric_limits<T>::min();
        Array<T,3> minPos(minValue,minValue,minValue), maxPos(maxValue,maxValue,maxValue);
        for (plint iVertex=0; iVertex<mesh.getNumVertices(); ++iVertex) {
            T nextValue = scalarData[iScalar][iVertex];
            if (nextValue<minValue) minValue = nextValue;
            if (nextValue>maxValue) maxValue = nextValue;
        }
        T range = maxValue-minValue;
        std::stringstream prop_str;
        prop_str <<  "            coloredProperties[\'" << scalars[iScalar] << "\']";
        std::string property = prop_str.str();
        ofile << property << " = new Object();\n";
        ofile << property << "[\'title\'] = \"" << "Palabos " << scalars[iScalar] << "\";\n";
        ofile << property << "[\'min\'] = " << minValue << ";\n";
        ofile << property << "[\'max\'] = " << maxValue << ";\n";
        ofile << property << "[\'value\'] = new Array(";
        for (plint iVertex=0; iVertex<mesh.getNumVertices(); ++iVertex) {
            T nextValue = (scalarData[iScalar][iVertex]-minValue)/range;
            ofile << nextValue;
            if (iVertex<mesh.getNumVertices()-1) {
                ofile << ", ";
            }
        }
        ofile << ");\n\n";
    }
    ofile << "        </script>\n";
    ofile << "    </head>\n";
    ofile << "    <x3d style=\"width: 100%; height :100%\">\n";
    ofile << "        <scene>\n";
    ofile << "            <shape>\n";
    ofile << "            <indexedtriangleset index=\"";

    for (plint iTriangle=0; iTriangle<mesh.getNumTriangles(); ++iTriangle) {
        plint i0 = mesh.getVertexId(iTriangle, 0);
        plint i1 = mesh.getVertexId(iTriangle, 1);
        plint i2 = mesh.getVertexId(iTriangle, 2);
        if ( tag<0 || boundary.getTag(iTriangle)==tag )
        {
            ofile << i0 << " " << i1 << " " << i2 << " ";
        }
    }
    ofile << "\" solid=\"false\">\n";
    ofile << "            <coordinate point=\"";
    for (plint iVertex=0; iVertex<mesh.getNumVertices(); ++iVertex) {
        ofile << posVect[iVertex][0] << " "
              << posVect[iVertex][1] << " "
              << posVect[iVertex][2];
        if (iVertex < mesh.getNumVertices()-1) {
            ofile << " ";
        }
    }
    ofile << "\"> </coordinate>\n";
    ofile << "            </indexedtriangleset>\n";
    ofile << "            </shape>\n";
    ofile << "            <viewpoint position=\'0.0 0.0 4.0\' orientation=\'0.0 0.0 0.0 0.0\'></viewpoint>\n";
    ofile << "        </scene>\n";
    ofile << "    </x3d>\n";
    ofile << "</html>\n";

    // Restore mesh selection which was active before calling
    //   this function.
    boundary.popSelect();
}

template<typename T, template<typename U> class Descriptor>
void exportX3dForVertices(std::vector<Particle3D<T,Descriptor>*> const& particles,
                          TriangleBoundary3D<T> const& boundary, std::vector<std::string> const& scalars,
                          bool dynamicMesh, plint tag, std::vector<T> const& scalarFactor, FileName fileName)
{
    if (boundary.getMesh().getNumTriangles() == 0) {
        return;
    }
    PLB_ASSERT( scalarFactor.empty() || scalarFactor.size()==scalars.size() );
    if (dynamicMesh) {
        boundary.pushSelect(0,1); // 0=Open, 1=Dynamic.
    }
    else {
        boundary.pushSelect(0,0); // Open, Static.
    }
    TriangularSurfaceMesh<T> const& mesh = boundary.getMesh();
    // If this assertion fails, a likely explanation is that the margin of your sparse block
    // structure is too small, and one of the particles was outside the allocated domain.
    // PLB_PRECONDITION((plint)particles.size() == mesh.getNumVertices());
    if((plint)particles.size() != mesh.getNumVertices()) {
        pcout << "Warning: in exportX3dForVertices, the numer of particles doesn't match the number" << std::endl;
        pcout << "of vertices. There might be black spots in the produced X3D file." << std::endl;
    }

    std::vector<std::vector<T> > scalarData(scalars.size());
    for (pluint iScalar=0; iScalar<scalars.size(); ++iScalar) {
        scalarData[iScalar].resize(mesh.getNumVertices());
    }
    std::vector<Array<T,3> > posVect(mesh.getNumVertices());
    // Default initialize all data, just in case every vertex hasn't
    // an associated particle.
    for (plint i=0; i<mesh.getNumVertices(); ++i) {
        for (pluint iScalar=0; iScalar<scalars.size(); ++iScalar) {
            scalarData[iScalar][i] = T();
        }
        posVect[i].resetToZero();
    }
    T maxLimit = std::numeric_limits<T>::max();
    T minLimit = std::numeric_limits<T>::min();
    Array<T,3> minPos(maxLimit,maxLimit,maxLimit), maxPos(minLimit,minLimit,minLimit);
    for (pluint iParticle=0; iParticle<particles.size(); ++iParticle) {
        plint iVertex = particles[iParticle]->getTag();
        posVect[iVertex] = particles[iParticle]->getPosition();
        if (posVect[iVertex][0]<minPos[0]) minPos[0] = posVect[iVertex][0];
        if (posVect[iVertex][1]<minPos[1]) minPos[1] = posVect[iVertex][1];
        if (posVect[iVertex][2]<minPos[2]) minPos[2] = posVect[iVertex][2];
        if (posVect[iVertex][0]>maxPos[0]) maxPos[0] = posVect[iVertex][0];
        if (posVect[iVertex][1]>maxPos[1]) maxPos[1] = posVect[iVertex][1];
        if (posVect[iVertex][2]>maxPos[2]) maxPos[2] = posVect[iVertex][2];
        for (pluint iScalar=0; iScalar<scalars.size(); ++iScalar) {
            T scalar;
            particles[iParticle]->getScalar(iScalar, scalar);
            if (!scalarFactor.empty()) {
                scalar *= scalarFactor[iScalar];
            }
            scalarData[iScalar][iVertex] = scalar;
        }
    }
    T diffx = maxPos[0]-minPos[0];
    T diffy = maxPos[1]-minPos[1];
    T diffz = maxPos[2]-minPos[2];
    T diff = std::max(diffx,std::max(diffy,diffz));
    T dx_nonDim = (T)2./diff;
    Array<T,3> offset_nonDim (
                 -((T)1.0+minPos[0]/diff),
                 -((T)1.0+minPos[1]/diff),
                 -((T)1.0+minPos[2]/diff) );
    for (pluint iParticle=0; iParticle<particles.size(); ++iParticle) {
        plint iVertex = particles[iParticle]->getTag();
        posVect[iVertex] *= dx_nonDim;
        posVect[iVertex] += offset_nonDim;
    }
    T dx_ndToPhys = boundary.getDx()/dx_nonDim;
    Array<T,3> location_ndToPhys = boundary.getPhysicalLocation()-dx_ndToPhys*offset_nonDim;
    
    std::ofstream coloredProperties_js(fileName.setName("coloredProperties.js").get().c_str());
    std::ofstream coordindex_dat      (fileName.setName("coordindex.dat").get().c_str());
    std::ofstream coordpoint_dat      (fileName.setName("coordpoint.dat").get().c_str());
    std::ofstream normalindex_dat     (fileName.setName("normalindex.dat").get().c_str());
    std::ofstream normalvector_dat    (fileName.setName("normalvector.dat").get().c_str());

    coloredProperties_js << "        var scale = " << std::setprecision(5) << dx_ndToPhys << ";\n";
    coloredProperties_js << "        var offset = new Array(" << std::setprecision(5)
                         << location_ndToPhys[0] << ","
                         << location_ndToPhys[1] << ","
                         << location_ndToPhys[2] << ");\n";
    coloredProperties_js << "        var coloredProperties = new Object();\n";

    for (pluint iScalar=0; iScalar<scalars.size(); ++iScalar) {
        T minValue = std::numeric_limits<T>::max();
        T maxValue = std::numeric_limits<T>::min();
        Array<T,3> minPos(minValue,minValue,minValue), maxPos(maxValue,maxValue,maxValue);
        for (plint iVertex=0; iVertex<mesh.getNumVertices(); ++iVertex) {
            T nextValue = scalarData[iScalar][iVertex];
            if (nextValue<minValue) minValue = nextValue;
            if (nextValue>maxValue) maxValue = nextValue;
        }
        T range = maxValue-minValue;
        std::stringstream prop_str;
        prop_str <<  "            coloredProperties[\'" << scalars[iScalar] << "\']";
        std::string property = prop_str.str();
        coloredProperties_js << property << " = new Object();\n";
        coloredProperties_js << property << "[\'title\'] = \"" << "Palabos " << scalars[iScalar] << "\";\n";
        coloredProperties_js << property << "[\'min\'] = " << minValue << ";\n";
        coloredProperties_js << property << "[\'max\'] = " << maxValue << ";\n";
        coloredProperties_js << property << "[\'value\'] = new Array(";
        for (plint iVertex=0; iVertex<mesh.getNumVertices(); ++iVertex) {
            T nextValue = (scalarData[iScalar][iVertex]-minValue)/range;
            coloredProperties_js << nextValue;
            if (iVertex<mesh.getNumVertices()-1) {
                coloredProperties_js << ", ";
            }
        }
        coloredProperties_js << ");\n";
    }

    for (plint iTriangle=0; iTriangle<mesh.getNumTriangles(); ++iTriangle) {
        plint i0 = mesh.getVertexId(iTriangle, 0);
        plint i1 = mesh.getVertexId(iTriangle, 1);
        plint i2 = mesh.getVertexId(iTriangle, 2);
        if ( tag<0 || boundary.getTag(iTriangle)==tag )
        {
            coordindex_dat << i0 << " " << i1 << " " << i2 << " ";
        }
    }

    for (plint iVertex=0; iVertex<mesh.getNumVertices(); ++iVertex) {
        normalindex_dat << iVertex;
        Array<T,3> normal = mesh.computeVertexNormal(iVertex);
        normalvector_dat << normal[0] << " "
                         << normal[1] << " "
                         << normal[2];
        coordpoint_dat << posVect[iVertex][0] << " "
                       << posVect[iVertex][1] << " "
                       << posVect[iVertex][2];
        if (iVertex < mesh.getNumVertices()-1) {
            coordpoint_dat << " ";
            normalindex_dat << " ";
            normalvector_dat << " ";
        }
    }
    // Restore mesh selection which was active before calling
    //   this function.
    boundary.popSelect();
}

template<typename T, template<typename U> class Descriptor>
void exportX3dMultiPart(TriangleBoundary3D<T> const& boundary, FileName fileName)
{
    if (!global::mpi().isMainProcessor()) {
        return;
    }
    if (boundary.getMesh().getNumTriangles() == 0) {
        return;
    }
    std::vector<Array<T,3> > posVect(boundary.getNumVertices());
    T maxLimit = std::numeric_limits<T>::max();
    T minLimit = std::numeric_limits<T>::min();
    Array<T,3> minPos(maxLimit,maxLimit,maxLimit), maxPos(minLimit,minLimit,minLimit);
    for (plint i=0; i<boundary.getNumVertices(); ++i) {
        posVect[i] = boundary.vertices()[i];
        if (posVect[i][0]<minPos[0]) minPos[0] = posVect[i][0];
        if (posVect[i][1]<minPos[1]) minPos[1] = posVect[i][1];
        if (posVect[i][2]<minPos[2]) minPos[2] = posVect[i][2];
        if (posVect[i][0]>maxPos[0]) maxPos[0] = posVect[i][0];
        if (posVect[i][1]>maxPos[1]) maxPos[1] = posVect[i][1];
        if (posVect[i][2]>maxPos[2]) maxPos[2] = posVect[i][2];
    }

    T diffx = maxPos[0]-minPos[0];
    T diffy = maxPos[1]-minPos[1];
    T diffz = maxPos[2]-minPos[2];
    T diff = std::max(diffx,std::max(diffy,diffz));
    T dx_nonDim = (T)2./diff;
    Array<T,3> offset_nonDim (
                 -((T)1.0+minPos[0]/diff),
                 -((T)1.0+minPos[1]/diff),
                 -((T)1.0+minPos[2]/diff) );
    for (plint i=0; i<boundary.getNumVertices(); ++i) {
        posVect[i] *= dx_nonDim;
        posVect[i] += offset_nonDim;
    }
    T phys_dx = boundary.getDx();
    Array<T,3> phys_location = boundary.getPhysicalLocation();
    T dx_ndToPhys = phys_dx/dx_nonDim;
    Array<T,3> location_ndToPhys = phys_location-dx_ndToPhys*offset_nonDim;

    std::ofstream coloredProperties_js(fileName.setName("coloredProperties.js").get().c_str());
    std::ofstream coordindex_dat      (fileName.setName("coordindex.dat").get().c_str());
    std::ofstream coordpoint_dat      (fileName.setName("coordpoint.dat").get().c_str());
    std::ofstream normalindex_dat     (fileName.setName("normalindex.dat").get().c_str());
    std::ofstream normalvector_dat    (fileName.setName("normalvector.dat").get().c_str());

    coloredProperties_js << "        var scale = " << std::setprecision(5) << dx_ndToPhys << ";\n";
    coloredProperties_js << "        var offset = new Array(" << std::setprecision(5)
                         << location_ndToPhys[0] << ","
                         << location_ndToPhys[1] << ","
                         << location_ndToPhys[2] << ");\n";
    
    for (plint iTriangle=0; iTriangle<boundary.getNumTriangles(); ++iTriangle) {
        plint i0 = boundary.getVertexId(iTriangle, 0);
        plint i1 = boundary.getVertexId(iTriangle, 1);
        plint i2 = boundary.getVertexId(iTriangle, 2);
        coordindex_dat << i0 << " " << i1 << " " << i2 << " ";
    }

    for (plint iVertex=0; iVertex<boundary.getNumVertices(); ++iVertex) {
        normalindex_dat << iVertex;
        Array<T,3> normal = boundary.computeVertexNormal(iVertex);
        normalvector_dat << normal[0] << " "
                         << normal[1] << " "
                         << normal[2];
        coordpoint_dat << posVect[iVertex][0] << " "
                       << posVect[iVertex][1] << " "
                       << posVect[iVertex][2];
        if (iVertex < boundary.getNumVertices()-1) {
            coordpoint_dat << " ";
            normalindex_dat << " ";
            normalvector_dat << " ";
        }
    }
}

}  // namespace plb

#endif  // PARTICLE_X3D_3D_HH

