/* This file is part of the Palabos library.
 *
 * Copyright (C) 2011-2014 FlowKit Sarl
 * Route d'Oron 2
 * 1010 Lausanne, Switzerland
 * E-mail contact: contact@flowkit.com
 *
 * The most recent release of Palabos can be downloaded at 
 * <http://www.palabos.org/>
 *
 * The library Palabos is free software: you can redistribute it and/or
 * modify it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * The library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef STREAMLINE_PARTICLE_3D_H
#define STREAMLINE_PARTICLE_3D_H

#include "core/globalDefs.h"
#include "core/array.h"
#include "core/geometry3D.h"
#include "particles/particle3D.h"
#include "particles/particleIdentifiers3D.h"
#include "algorithm/spline.h"
#include "offLattice/triangleSet.h"

#include <vector>

namespace plb {

/* StreamlineParticle3D: This particle type, is used to compute streamlines.
 * It stores all positions that it passes from during its existence,
 * so the streamline can later be plotted. There is the option of
 * "useConstantSpatialStepIntegration". With this, the equations of motion
 * become: dx/ds = u*lamda, instead of dx/ds = u. The parameter "lamda" is
 * such that norm(u)*lamda*ds = constantSpatialStep, with "costantSpatialStep"
 * provided by the user. ds, which is different from dt, is referred to as
 * "streamlineStep". The above integration scheme is used for efficiency reasons,
 * but it might deteriorate accuracy. There is also the notion of the "domain".
 * Streamline particles that pass beyond the boundaries of the "domain", are
 * "frozen" and their advancement stops. The recommended use of the "domain"
 * parameter is the following. The user defines "domain" to be the bounding box
 * of the streamlines. He should construct a velocity field from the simulation lattice,
 * only for this domain. Then the streamlines will be computed only inside this
 * specific box. Naturally, all streamline seeding points must be contained in the
 * "domain" as well. Technically, the bounding box of the streamlines is
 * "domain.enlarge(-1)", so the user must not seed the streamline particles
 * very close to the boundaries of the "domain". The streamlines are never constructed
 * to be periodic, even if the velocity field is. Needless to say that the "domain"
 * must be at most as big as the bounding box of the simulation lattice (even for
 * periodic problems).
 * This is an abstract class since the function "createStreamline" is not
 * implemented here.
 * */
template<typename T, template<typename U> class Descriptor>
class StreamlineParticle3D : public PointParticle3D<T,Descriptor> {
public:
    StreamlineParticle3D();
    // Constructor for a StreamlineParticle3D which behaves exactly as a PointParticle3D.
    StreamlineParticle3D(plint tag_, Array<T,3> const& position_, Array<T,3> const& velocity_,
            Box3D const& domain_, bool useFrenetSerret_=false);
    // If useConstantSpatialStepIntegration_ is true, then step_ is the constantSpatialStep,
    // otherwise it is the streamlineStep.
    StreamlineParticle3D(plint tag_, Array<T,3> const& position_, Array<T,3> const& velocity_,
            bool useConstantSpatialStepIntegration_, T step_, Box3D const& domain_,
            bool useFrenetSerret_=false);
    virtual void advance();
    virtual void serialize(HierarchicSerializer& serializer) const;
    virtual void unserialize(HierarchicUnserializer& unserializer);
    virtual void rescale(int dxScale, int dtScale);
    std::vector<Array<T,3> > const& getPositions() const { return positions; }
    std::vector<Array<T,3> >& getPositions() { return positions; }
    bool hasStoppedAdvancing() const { return stoppedAdvancing; }
    void toggleConstantSpatialStepIntegration(bool onOff) { useConstantSpatialStepIntegration = onOff; }
    void toggleFrenetSerret(bool onOff) { useFrenetSerret = onOff; }
    T& getConstantSpatialStep() { return constantSpatialStep; }
    T getConstantSpatialStep() const { return constantSpatialStep; }
    T& getStreamlineStep() { return streamlineStep; }
    T getStreamlineStep() const { return streamlineStep; }
    Box3D& getDomain() { return domain; }
    Box3D getDomain() const { return domain; }
    Precision getPrecision() const { return precision; }
    // Uniformly resample the position and velocity containers. The new ones will have a size equal to "n".
    // Returns true if the resampling proceedure was successful, of false otherwise.
    bool resample(plint n);
    // Creates a streamline as a triangle set.
    virtual void createStreamline(TriangleSet<T>& streamline) const = 0;
protected:
    void constructSplineVectorRepresentation(std::vector<T> const& scalarParameter,
            std::vector<Array<T,3> > const& vectorFunction, NaturalCubicSpline<T>** component0,
            NaturalCubicSpline<T>** component1, NaturalCubicSpline<T>** component2) const;
    void destructSplineVectorRepresentation(NaturalCubicSpline<T>* component0, NaturalCubicSpline<T>* component1,
            NaturalCubicSpline<T>* component2) const;
    int constructOrthonormalFrames(std::vector<T> const& t, NaturalCubicSpline<T>* x, NaturalCubicSpline<T>* y,
            NaturalCubicSpline<T>* z, std::vector<Array<T,3> >& tangents, std::vector<Array<T,3> >& normals,
            std::vector<Array<T,3> >& binormals) const;
private:
    std::vector<Array<T,3> > positions;
    bool stoppedAdvancing;
    bool useConstantSpatialStepIntegration;
    T constantSpatialStep;
    T streamlineStep;
    Box3D domain;
    bool useFrenetSerret;
    Precision precision;
};

/* SphereStreamlineParticle3D: This particle plots streamlines as distinct spheres.
 * Each sphere has a radius "radius", and is computed with at least "minNumOfTriangles"
 * triangles.
 * */
template<typename T, template<typename U> class Descriptor>
class SphereStreamlineParticle3D : public StreamlineParticle3D<T,Descriptor> {
public:
    SphereStreamlineParticle3D();
    SphereStreamlineParticle3D(plint tag_, Array<T,3> const& position_, Array<T,3> const& velocity_,
            Box3D const& domain_, T radius_, plint minNumOfTriangles_);
    SphereStreamlineParticle3D(plint tag_, Array<T,3> const& position_, Array<T,3> const& velocity_,
            bool useConstantSpatialStepIntegration_, T step_, Box3D const& domain_, T radius_,
            plint minNumOfTriangles_);
    virtual SphereStreamlineParticle3D<T,Descriptor>* clone() const;
    virtual int getId() const;
    virtual void serialize(HierarchicSerializer& serializer) const;
    virtual void unserialize(HierarchicUnserializer& unserializer);
    virtual void rescale(int dxScale, int dtScale);
    T getRadius() const { return radius; }
    T& getRadius() { return radius; }
    plint getMinNumOfTriangles() const { return minNumOfTriangles; }
    plint& getMinNumOfTriangles() { return minNumOfTriangles; }
    virtual void createStreamline(TriangleSet<T>& streamline) const;
private:
    T radius;
    plint minNumOfTriangles;
    static int id;
};

/* TubeStreamlineParticle3D: This particle plots streamlines as cylindrical tubes
 * with an elliptical cross-section. Each cross-section is computed with "numPoints"
 * points in the circumferential direction, and has the shape of an ellipse with radii
 * "normalRadius" and "binormalRadius" in the directions of the normal and binormal
 * vectors of the orthonormal frame.
 * */
template<typename T, template<typename U> class Descriptor>
class TubeStreamlineParticle3D : public StreamlineParticle3D<T,Descriptor> {
public:
    TubeStreamlineParticle3D();
    TubeStreamlineParticle3D(plint tag_, Array<T,3> const& position_, Array<T,3> const& velocity_,
            Box3D const& domain_, T normalRadius_, T binormalRadius_, plint numPoints_,
            bool useFrenetSerret_=false);
    TubeStreamlineParticle3D(plint tag_, Array<T,3> const& position_, Array<T,3> const& velocity_,
            bool useConstantSpatialStepIntegration_, T step_, Box3D const& domain_,
            T normalRadius_, T binormalRadius_, plint numPoints_, bool useFrenetSerret_=false);
    virtual TubeStreamlineParticle3D<T,Descriptor>* clone() const;
    virtual int getId() const;
    virtual void serialize(HierarchicSerializer& serializer) const;
    virtual void unserialize(HierarchicUnserializer& unserializer);
    virtual void rescale(int dxScale, int dtScale);
    T getNormalRadius() const { return normalRadius; }
    T& getNormalRadius() { return normalRadius; }
    T getBinormalRadius() const { return binormalRadius; }
    T& getBinormalRadius() { return binormalRadius; }
    plint getNumPoints() const { return numPoints; }
    plint& getNumPoints() { return numPoints; }
    virtual void createStreamline(TriangleSet<T>& streamline) const;
private:
    T normalRadius;
    T binormalRadius;
    plint numPoints;
    static int id;
};

/* StripStreamlineParticle3D: This particle plots streamlines as strips of thickness
 * "thickness". The direction of the strip lies 45 degrees in-between the normal
 * and the binormal vectors of the orthonormal frame.
 * */
template<typename T, template<typename U> class Descriptor>
class StripStreamlineParticle3D : public StreamlineParticle3D<T,Descriptor> {
public:
    StripStreamlineParticle3D();
    StripStreamlineParticle3D(plint tag_, Array<T,3> const& position_, Array<T,3> const& velocity_,
            Box3D const& domain_, T thickness_, bool useFrenetSerret_=false);
    StripStreamlineParticle3D(plint tag_, Array<T,3> const& position_, Array<T,3> const& velocity_,
            bool useConstantSpatialStepIntegration_, T step_, Box3D const& domain_,
            T thickness_, bool useFrenetSerret_=false);
    virtual StripStreamlineParticle3D<T,Descriptor>* clone() const;
    virtual int getId() const;
    virtual void serialize(HierarchicSerializer& serializer) const;
    virtual void unserialize(HierarchicUnserializer& unserializer);
    virtual void rescale(int dxScale, int dtScale);
    T getThickness() const { return thickness; }
    T& getThickness() { return thickness; }
    virtual void createStreamline(TriangleSet<T>& streamline) const;
private:
    T thickness;
    static int id;
};

}  // namespace plb

#endif  // STREAMLINE_PARTICLE_3D_H
