/* This file is part of the Palabos library.
 *
 * Copyright (C) 2011-2014 FlowKit Sarl
 * Route d'Oron 2
 * 1010 Lausanne, Switzerland
 * E-mail contact: contact@flowkit.com
 *
 * The most recent release of Palabos can be downloaded at 
 * <http://www.palabos.org/>
 *
 * The library Palabos is free software: you can redistribute it and/or
 * modify it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * The library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef STREAMLINE_PARTICLE_3D_HH
#define STREAMLINE_PARTICLE_3D_HH

#include "core/globalDefs.h"
#include "latticeBoltzmann/geometricOperationTemplates.h"
#include "geometry/streamlineParticle3D.h"
#include "offLattice/triangleSetGenerator.h"
#include "multiGrid/multiGridUtil.h"

#include <limits>
#include <cmath>

namespace plb {

/* *************** class StreamlineParticle3D ************************************ */

template<typename T, template<typename U> class Descriptor>
StreamlineParticle3D<T,Descriptor>::StreamlineParticle3D()
    : PointParticle3D<T,Descriptor>(),
      stoppedAdvancing(false),
      useConstantSpatialStepIntegration(false),
      constantSpatialStep((T) 0),
      streamlineStep((T) 1),
      useFrenetSerret(false)
{
    domain = Box3D(0, 0, 0, 0, 0, 0);
    precision = floatingPointPrecision<T>();
}

template<typename T, template<typename U> class Descriptor>
StreamlineParticle3D<T,Descriptor>::StreamlineParticle3D(plint tag_, Array<T,3> const& position_,
        Array<T,3> const& velocity_, Box3D const& domain_, bool useFrenetSerret_)
    : PointParticle3D<T,Descriptor>(tag_, position_, velocity_),
      stoppedAdvancing(false),
      useConstantSpatialStepIntegration(false),
      constantSpatialStep((T) 0),
      streamlineStep((T) 1),
      domain(domain_),
      useFrenetSerret(useFrenetSerret_)
{
    precision = floatingPointPrecision<T>();
    PLB_ASSERT(domain.x1-domain.x0 >= 3 && domain.y1-domain.y0 >= 3 && domain.z1-domain.z0 >= 3);
}

template<typename T, template<typename U> class Descriptor>
StreamlineParticle3D<T,Descriptor>::StreamlineParticle3D(plint tag_, Array<T,3> const& position_,
        Array<T,3> const& velocity_, bool useConstantSpatialStepIntegration_, T step_,
        Box3D const& domain_, bool useFrenetSerret_)
    : PointParticle3D<T,Descriptor>(tag_, position_, velocity_),
      stoppedAdvancing(false),
      useConstantSpatialStepIntegration(useConstantSpatialStepIntegration_),
      domain(domain_),
      useFrenetSerret(useFrenetSerret_)
{
    precision = floatingPointPrecision<T>();
#ifdef PLB_DEBUG
    T eps = getEpsilon<T>(precision);
#endif

    if (useConstantSpatialStepIntegration) {
        constantSpatialStep = step_;
        PLB_ASSERT(std::fabs(constantSpatialStep) > eps);
        PLB_ASSERT(std::fabs(constantSpatialStep) < (T) 1);
        streamlineStep = 1.0;
    } else {
        streamlineStep = step_;
        PLB_ASSERT(std::fabs(streamlineStep) > eps);
        constantSpatialStep = 0.0;
    }

    PLB_ASSERT(domain.x1-domain.x0 >= 3 && domain.y1-domain.y0 >= 3 && domain.z1-domain.z0 >= 3);
}

template<typename T, template<typename U> class Descriptor>
void StreamlineParticle3D<T,Descriptor>::advance()
{
    static T eps = getEpsilon<T>(precision);

    if (!stoppedAdvancing) {
        Array<T,3> velocity = this->getVelocity();
        T velocityNorm = norm<T,3>(velocity);
        if (velocityNorm <= eps) {
            stoppedAdvancing = true;
            return;
        }

        Array<T,3> oldPosition = this->getPosition();
        if (!contained(oldPosition, domain.enlarge(-1))) {
            stoppedAdvancing = true;
            return;
        }

        positions.push_back(oldPosition);
        Array<T,3> scaledVelocity(velocity);
        if (useConstantSpatialStepIntegration) {
            T lamda = constantSpatialStep / (streamlineStep * velocityNorm);
            scaledVelocity *= lamda;
        }
        this->getPosition() += streamlineStep * scaledVelocity;
    }
}

template<typename T, template<typename U> class Descriptor>
void StreamlineParticle3D<T,Descriptor>::serialize(HierarchicSerializer& serializer) const
{
    PointParticle3D<T,Descriptor>::serialize(serializer);
    pluint n = (pluint) positions.size();
    serializer.addValue(n);
    for (pluint i = 0; i < positions.size(); i++) {
        serializer.addValues<T,3>(positions[i]);
    }
    serializer.addValue(stoppedAdvancing);
    serializer.addValue(useConstantSpatialStepIntegration);
    serializer.addValue(constantSpatialStep);
    serializer.addValue(streamlineStep);
    serializer.addValue(domain.x0);
    serializer.addValue(domain.x1);
    serializer.addValue(domain.y0);
    serializer.addValue(domain.y1);
    serializer.addValue(domain.z0);
    serializer.addValue(domain.z1);
    serializer.addValue(useFrenetSerret);
    serializer.addValue(precision);
}

template<typename T, template<typename U> class Descriptor>
void StreamlineParticle3D<T,Descriptor>::unserialize(HierarchicUnserializer& unserializer)
{
    PointParticle3D<T,Descriptor>::unserialize(unserializer);
    pluint n;
    unserializer.readValue(n);
    positions.clear();
    positions.resize(n);
    for (pluint i = 0; i < positions.size(); i++) {
        unserializer.readValues<T,3>(positions[i]);
    }
    unserializer.readValue(stoppedAdvancing);
    unserializer.readValue(useConstantSpatialStepIntegration);
    unserializer.readValue(constantSpatialStep);
    unserializer.readValue(streamlineStep);
    unserializer.readValue(domain.x0);
    unserializer.readValue(domain.x1);
    unserializer.readValue(domain.y0);
    unserializer.readValue(domain.y1);
    unserializer.readValue(domain.z0);
    unserializer.readValue(domain.z1);
    unserializer.readValue(useFrenetSerret);
    unserializer.readValue(precision);
}

template<typename T, template<typename U> class Descriptor>
void StreamlineParticle3D<T,Descriptor>::rescale(int dxScale, int dtScale)
{
    PointParticle3D<T,Descriptor>::rescale(dxScale, dtScale);

    T posDimDx = 1.0;
    T posDimDt = 0.0;
    T posScaleFactor = scaleFromReference(dxScale, posDimDx, dtScale, posDimDt);
    for (pluint i = 0; i < positions.size(); i++) {
        positions[i] *= posScaleFactor;
    }
    constantSpatialStep *= posScaleFactor;

    domain.x0 *= posScaleFactor;
    domain.x1 *= posScaleFactor;
    domain.y0 *= posScaleFactor;
    domain.y1 *= posScaleFactor;
    domain.z0 *= posScaleFactor;
    domain.z1 *= posScaleFactor;

    T timeDimDx = 0.0;
    T timeDimDt = 1.0;
    T timeScaleFactor = scaleFromReference(dxScale, timeDimDx, dtScale, timeDimDt);
    streamlineStep *= timeScaleFactor;
}

template<typename T, template<typename U> class Descriptor>
bool StreamlineParticle3D<T,Descriptor>::resample(plint n)
{
    plint size = (plint) positions.size();

    if (n < 0) {
        return false;
    }

    // n >= 0 from now on...
    if (n == 0) {
        positions.clear();
        return true;
    }

    // n >= 1 from now on...
    if (size == 0) {
        return false;
    }

    // size >= 1 from now on...
    if (n == 1) {
        Array<T,3> position(positions[0]);
        positions.clear();
        positions.resize(1);
        positions[0] = position;
        return true;
    }

    // n >= 2 from now on...
    if (size == 1) {
        return false;
    }

    // size >= 2 from now on...
    if (n == 2) {
        Array<T,3> position0(positions[0]);
        Array<T,3> position1(positions[size-1]);
        positions.clear();
        positions.resize(2);
        positions[0] = position0;
        positions[1] = position1;
        return true;
    }

    // n >= 3 and size >= 2 from now on...
    // Spline interpolation is needed.

    // The spline interpolation parametrizes the functions with respect to the index in the container.
    std::vector<T> t(size);
    for (plint i = 0; i < size; i++) {
        t[i] = (T) i;
    }
    T dt = (t[size-1] - t[0]) / (T) (n - 1);

    NaturalCubicSpline<T> *x, *y, *z;
    constructSplineVectorRepresentation(t, positions, &x, &y, &z);
    positions.clear();
    positions.resize(n);
    for (plint i = 0; i < n; i++) {
        T tSample = t[0] + (T) i * dt;
        positions[i] = Array<T,3>(x->getFunctionValue(tSample), y->getFunctionValue(tSample), z->getFunctionValue(tSample));
    }
    destructSplineVectorRepresentation(x, y, z);

    return true;
}

template<typename T, template<typename U> class Descriptor>
void StreamlineParticle3D<T,Descriptor>::constructSplineVectorRepresentation(std::vector<T> const& scalarParameter,
        std::vector<Array<T,3> > const& vectorFunction, NaturalCubicSpline<T>** component0,
        NaturalCubicSpline<T>** component1, NaturalCubicSpline<T>** component2) const
{
    PLB_ASSERT(scalarParameter.size() == vectorFunction.size());
    plint size = (plint) scalarParameter.size();

    std::vector<T> buf(size);

    for (plint i = 0; i < size; i++) {
        buf[i] = vectorFunction[i][0];
    }
    *component0 = new NaturalCubicSpline<T>(scalarParameter, buf);
    for (plint i = 0; i < size; i++) {
        buf[i] = vectorFunction[i][1];
    }
    *component1 = new NaturalCubicSpline<T>(scalarParameter, buf);
    for (plint i = 0; i < size; i++) {
        buf[i] = vectorFunction[i][2];
    }
    *component2 = new NaturalCubicSpline<T>(scalarParameter, buf);
}

template<typename T, template<typename U> class Descriptor>
void StreamlineParticle3D<T,Descriptor>::destructSplineVectorRepresentation(NaturalCubicSpline<T>* component0,
        NaturalCubicSpline<T>* component1, NaturalCubicSpline<T>* component2) const
{
    delete component0;
    delete component1;
    delete component2;
}

template<typename T, template<typename U> class Descriptor>
int StreamlineParticle3D<T,Descriptor>::constructOrthonormalFrames(std::vector<T> const& t,
        NaturalCubicSpline<T>* x, NaturalCubicSpline<T>* y, NaturalCubicSpline<T>* z,
        std::vector<Array<T,3> >& tangents, std::vector<Array<T,3> >& normals,
        std::vector<Array<T,3> >& binormals) const
{
    //static T pi = std::acos((T) -1.0);
    static T eps = getEpsilon<T>(precision);
    int error = 0;

    if (t.size() <= 1) {
        tangents.clear();
        normals.clear();
        binormals.clear();
        error = 1;
        return error;
    }
    plint size = (plint) t.size();

    tangents.clear();
    normals.clear();
    binormals.clear();

    tangents.resize(size);
    normals.resize(size);
    binormals.resize(size);

    if (useFrenetSerret) {
        // Frenet-Serret orthonormal frame. It does not always give the best results.
        for (plint i = 0; i < size; i++) {
            T dx = x->getDerivativeValue(t[i]);
            T dy = y->getDerivativeValue(t[i]);
            T dz = z->getDerivativeValue(t[i]);

            T l = std::sqrt(dx*dx + dy*dy + dz*dz);
            if (l <= eps) {
                tangents.clear();
                normals.clear();
                binormals.clear();
                error = 1;
                return error;
            }

            tangents[i] = Array<T,3>(dx/l, dy/l, dz/l);

            T ddx = x->getSecondDerivativeValue(t[i]);
            T ddy = y->getSecondDerivativeValue(t[i]);
            T ddz = z->getSecondDerivativeValue(t[i]);

            T dl = (dx*ddx + dy*ddy + dz*ddz) / l;
            Array<T,3> dtangent = Array<T,3>((ddx*l-dx*dl)/(l*l), (ddy*l-dy*dl)/(l*l), (ddz*l-dz*dl)/(l*l));
            T dtangentNorm = norm<T,3>(dtangent);

            bool undefinedNormal = false;
            if (dtangentNorm <= eps) {
                undefinedNormal = true;
                plint j = 1;
                int numberOfTries = 10;
                while (undefinedNormal && (i-j >= 0 || i+j < size)) {
                    if (i-j >= 0) {
                        T t1 = t[i-j];
                        T t2 = t[i-j+1];
                        T dt = (t2 - t1) / (T) (numberOfTries - 1);
                        for (int iTry = 0; iTry < numberOfTries; iTry++) {
                            T tnew = t1 + (T) iTry * dt;
                            ddx = x->getSecondDerivativeValue(tnew);
                            ddy = y->getSecondDerivativeValue(tnew);
                            ddz = z->getSecondDerivativeValue(tnew);
                            dl = (dx*ddx + dy*ddy + dz*ddz) / l;
                            dtangent = Array<T,3>((ddx*l-dx*dl)/(l*l), (ddy*l-dy*dl)/(l*l), (ddz*l-dz*dl)/(l*l));
                            dtangentNorm = norm<T,3>(dtangent);
                            if (dtangentNorm > eps) {
                                undefinedNormal = false;
                                break;
                            }
                        }
                    }

                    if (undefinedNormal && (i+j < size)) {
                        T t1 = t[i+j-1];
                        T t2 = t[i+j];
                        T dt = (t2 - t1) / (T) (numberOfTries - 1);
                        for (int iTry = 0; iTry < numberOfTries; iTry++) {
                            T tnew = t1 + (T) iTry * dt;
                            ddx = x->getSecondDerivativeValue(tnew);
                            ddy = y->getSecondDerivativeValue(tnew);
                            ddz = z->getSecondDerivativeValue(tnew);
                            dl = (dx*ddx + dy*ddy + dz*ddz) / l;
                            dtangent = Array<T,3>((ddx*l-dx*dl)/(l*l), (ddy*l-dy*dl)/(l*l), (ddz*l-dz*dl)/(l*l));
                            dtangentNorm = norm<T,3>(dtangent);
                            if (dtangentNorm > eps) {
                                undefinedNormal = false;
                                break;
                            }
                        }
                    }
                    j++;
                }
            }
            if (undefinedNormal) {
                crossProduct<T>(tangents[i], Array<T,3>((T) 1, (T) 0, (T) 0), normals[i]);
                T normNormal = norm<T,3>(normals[i]);
                if (normNormal <= eps) {
                    crossProduct<T>(tangents[i], Array<T,3>((T) 0, (T) 1, (T) 0), normals[i]);
                    normNormal = norm<T,3>(normals[i]);
                    if (normNormal <= eps) {
                        crossProduct<T>(tangents[i], Array<T,3>((T) 0, (T) 0, (T) 1), normals[i]);
                        normNormal = norm<T,3>(normals[i]);
                        if (normNormal <= eps) {
                            tangents.clear();
                            normals.clear();
                            binormals.clear();
                            error = 1;
                            return error;
                        }
                    }
                }
                normals[i] /= normNormal;
            } else {
                normals[i] = dtangent / dtangentNorm;
            }

            crossProduct<T>(tangents[i], normals[i], binormals[i]);
        }
    } else {
        // Default orthonormal frame. Gives very reliable results.

        // Construct initial default frame.
        T dx = x->getDerivativeValue(t[0]);
        T dy = y->getDerivativeValue(t[0]);
        T dz = z->getDerivativeValue(t[0]);

        T l = std::sqrt(dx*dx + dy*dy + dz*dz);
        if (l <= eps) {
            tangents.clear();
            normals.clear();
            binormals.clear();
            error = 1;
            return error;
        }

        tangents[0] = Array<T,3>(dx/l, dy/l, dz/l);

        crossProduct<T>(tangents[0], Array<T,3>((T) 1, (T) 0, (T) 0), normals[0]);
        T normNormal = norm<T,3>(normals[0]);
        if (normNormal <= eps) {
            crossProduct<T>(tangents[0], Array<T,3>((T) 0, (T) 1, (T) 0), normals[0]);
            normNormal = norm<T,3>(normals[0]);
            if (normNormal <= eps) {
                crossProduct<T>(tangents[0], Array<T,3>((T) 0, (T) 0, (T) 1), normals[0]);
                normNormal = norm<T,3>(normals[0]);
                if (normNormal <= eps) {
                    tangents.clear();
                    normals.clear();
                    binormals.clear();
                    error = 1;
                    return error;
                }
            }
        }
        normals[0] /= normNormal;

        // Make sure the normal is normal to the tangent and of unitary length.
        normals[0] = normals[0] - dot<T,3>(tangents[0], normals[0]) * tangents[0];
        normNormal = norm<T,3>(normals[0]);
        PLB_ASSERT(normNormal > eps);
        normals[0] /= normNormal;

        crossProduct<T>(tangents[0], normals[0], binormals[0]);

        // Make sure the binormal is normal to the normal and of unitary length.
        binormals[0] = binormals[0] - dot<T,3>(normals[0], binormals[0]) * normals[0];
        T normBinormal = norm<T,3>(binormals[0]);
        PLB_ASSERT(normBinormal > eps);
        binormals[0] /= normBinormal;

        // For all next frames, rotate the previous normal and binormal, so that the
        // previous tangent coincides with the new one.
        Array<T,3> prevTangent(tangents[0]);
        Array<T,3> prevNormal(normals[0]);
        Array<T,3> prevBinormal(binormals[0]);
        for (plint i = 1; i < size; i++) {
            T dx = x->getDerivativeValue(t[i]);
            T dy = y->getDerivativeValue(t[i]);
            T dz = z->getDerivativeValue(t[i]);

            T l = std::sqrt(dx*dx + dy*dy + dz*dz);
            if (l <= eps) {
                tangents.clear();
                normals.clear();
                binormals.clear();
                error = 1;
                return error;
            }

            tangents[i] = Array<T,3>(dx/l, dy/l, dz/l);

            Array<T,3> tangent(tangents[i]);

            T theta = angleBetweenVectors<T>(tangent, prevTangent);
            Array<T,3> normedAxis = crossProduct<T>(prevTangent, tangent);

            // Make sure the normedAxis is normal to the tangent.
            normedAxis = normedAxis - dot<T,3>(tangent, normedAxis) * tangent;

            T r = norm<T,3>(normedAxis);
            if (r <= eps) {
                // If the cross-product is 0, then the dot-product must be 1,
                // meaning that tangent and prevTangent are the same.
                PLB_ASSERT(std::fabs(dot<T,3>(prevTangent, tangent) - (T) 1) <= eps);
                // TODO: In these cases although the local frame can be considered
                //       rotating around the tangent, here we consider that it does not
                //       rotate at all.

                tangents[i] = prevTangent;
                normals[i] = prevNormal;
                binormals[i] = prevBinormal;

                // TODO: OR?
                /*
                normedAxis = prevNormal;
                if (dot<T,3>(prevTangent, tangent) > 0.0) {
                    theta = 0.0;
                } else {
                    theta = pi;
                }
                normals[i] = rotateAtOrigin<T>(prevNormal, normedAxis, theta);
                binormals[i] = rotateAtOrigin<T>(prevBinormal, normedAxis, theta);

                prevTangent = tangents[i];
                prevNormal = normals[i];
                prevBinormal = binormals[i];
                */
            } else {
                normedAxis /= r;

                normals[i] = rotateAtOrigin<T>(prevNormal, normedAxis, theta);

                // Make sure the normal is normal to the tangent and of unitary length.
                normals[i] = normals[i] - dot<T,3>(tangents[i], normals[i]) * tangents[i];
                T normNormal = norm<T,3>(normals[i]);
                PLB_ASSERT(normNormal > eps);
                normals[i] /= normNormal;

                crossProduct<T>(tangents[i], normals[i], binormals[i]);

                // Make sure the binormal is normal to the normal and of unitary length.
                binormals[i] = binormals[i] - dot<T,3>(normals[i], binormals[i]) * normals[i];
                T normBinormal = norm<T,3>(binormals[i]);
                PLB_ASSERT(normBinormal > eps);
                binormals[i] /= normBinormal;

                //binormals[i] = rotateAtOrigin<T>(prevBinormal, normedAxis, theta);

                prevTangent = tangents[i];
                prevNormal = normals[i];
                prevBinormal = binormals[i];
            }
        }
    }

    return 0;
}


/* *************** class SphereStreamlineParticle3D ************************************ */

template<typename T, template<typename U> class Descriptor>
int SphereStreamlineParticle3D<T,Descriptor>::id =
        meta::registerGenericParticle3D<T,Descriptor,SphereStreamlineParticle3D<T,Descriptor> >("SphereStreamline");

template<typename T, template<typename U> class Descriptor>
SphereStreamlineParticle3D<T,Descriptor>::SphereStreamlineParticle3D()
    : StreamlineParticle3D<T,Descriptor>(),
      radius((T) 0),
      minNumOfTriangles(0)
{ }

template<typename T, template<typename U> class Descriptor>
SphereStreamlineParticle3D<T,Descriptor>::SphereStreamlineParticle3D(plint tag_, Array<T,3> const& position_,
        Array<T,3> const& velocity_, Box3D const& domain_, T radius_, plint minNumOfTriangles_)
    : StreamlineParticle3D<T,Descriptor>(tag_, position_, velocity_, domain_),
      radius(radius_),
      minNumOfTriangles(minNumOfTriangles_)
{
    T eps = getEpsilon<T>(this->getPrecision());
    PLB_ASSERT(radius > eps);
    if (minNumOfTriangles < 8) {
        minNumOfTriangles = 8;
    }
}

template<typename T, template<typename U> class Descriptor>
SphereStreamlineParticle3D<T,Descriptor>::SphereStreamlineParticle3D(plint tag_, Array<T,3> const& position_,
        Array<T,3> const& velocity_, bool useConstantSpatialStepIntegration_, T step_, Box3D const& domain_,
        T radius_, plint minNumOfTriangles_)
    : StreamlineParticle3D<T,Descriptor>(tag_, position_, velocity_, useConstantSpatialStepIntegration_,
            step_, domain_),
      radius(radius_),
      minNumOfTriangles(minNumOfTriangles_)
{
#ifdef PLB_DEBUG
    T eps = getEpsilon<T>(this->getPrecision());
#endif
    PLB_ASSERT(radius > eps);
    if (minNumOfTriangles < 8) {
        minNumOfTriangles = 8;
    }
}

template<typename T, template<typename U> class Descriptor>
SphereStreamlineParticle3D<T,Descriptor>* SphereStreamlineParticle3D<T,Descriptor>::clone() const
{
    return new SphereStreamlineParticle3D<T,Descriptor>(*this);
}

template<typename T, template<typename U> class Descriptor>
int SphereStreamlineParticle3D<T,Descriptor>::getId() const 
{
    return id;
}

template<typename T, template<typename U> class Descriptor>
void SphereStreamlineParticle3D<T,Descriptor>::serialize(HierarchicSerializer& serializer) const
{
    StreamlineParticle3D<T,Descriptor>::serialize(serializer);
    serializer.addValue(radius);
    serializer.addValue(minNumOfTriangles);
}

template<typename T, template<typename U> class Descriptor>
void SphereStreamlineParticle3D<T,Descriptor>::unserialize(HierarchicUnserializer& unserializer)
{
    StreamlineParticle3D<T,Descriptor>::unserialize(unserializer);
    unserializer.readValue(radius);
    unserializer.readValue(minNumOfTriangles);
}

template<typename T, template<typename U> class Descriptor>
void SphereStreamlineParticle3D<T,Descriptor>::rescale(int dxScale, int dtScale)
{
    StreamlineParticle3D<T,Descriptor>::rescale(dxScale, dtScale);

    T posDimDx = 1.0;
    T posDimDt = 0.0;
    T posScaleFactor = scaleFromReference(dxScale, posDimDx, dtScale, posDimDt);
    radius *= posScaleFactor;
}

template<typename T, template<typename U> class Descriptor>
void SphereStreamlineParticle3D<T,Descriptor>::createStreamline(TriangleSet<T>& streamline) const
{
    std::vector<Array<T,3> > const& positions = this->getPositions();
    streamline.clear();
    Precision precision = this->getPrecision();
    streamline.setPrecision(precision);
    if (positions.size() < 1) {
        return;
    }
    for (plint i = 0; i < (plint) positions.size(); i++) {
        TriangleSet<T> sphere = constructSphere<T>(positions[i], radius, minNumOfTriangles);
        streamline.append(sphere);
    }
}


/* *************** class TubeStreamlineParticle3D ************************************ */

template<typename T, template<typename U> class Descriptor>
int TubeStreamlineParticle3D<T,Descriptor>::id =
        meta::registerGenericParticle3D<T,Descriptor,TubeStreamlineParticle3D<T,Descriptor> >("TubeStreamline");

template<typename T, template<typename U> class Descriptor>
TubeStreamlineParticle3D<T,Descriptor>::TubeStreamlineParticle3D()
    : StreamlineParticle3D<T,Descriptor>(),
      normalRadius((T) 0),
      binormalRadius((T) 0),
      numPoints(0)
{ }

template<typename T, template<typename U> class Descriptor>
TubeStreamlineParticle3D<T,Descriptor>::TubeStreamlineParticle3D(plint tag_, Array<T,3> const& position_,
        Array<T,3> const& velocity_, Box3D const& domain_, T normalRadius_, T binormalRadius_,
        plint numPoints_, bool useFrenetSerret_)
    : StreamlineParticle3D<T,Descriptor>(tag_, position_, velocity_, domain_, useFrenetSerret_),
      normalRadius(normalRadius_),
      binormalRadius(binormalRadius_),
      numPoints(numPoints_)
{
    T eps = getEpsilon<T>(this->getPrecision());
    PLB_ASSERT(normalRadius > eps);
    PLB_ASSERT(binormalRadius > eps);
    if (numPoints < 3) {
        numPoints = 3;
    }
}

template<typename T, template<typename U> class Descriptor>
TubeStreamlineParticle3D<T,Descriptor>::TubeStreamlineParticle3D(plint tag_, Array<T,3> const& position_,
        Array<T,3> const& velocity_, bool useConstantSpatialStepIntegration_, T step_, Box3D const& domain_,
        T normalRadius_, T binormalRadius_, plint numPoints_, bool useFrenetSerret_)
    : StreamlineParticle3D<T,Descriptor>(tag_, position_, velocity_, useConstantSpatialStepIntegration_,
            step_, domain_, useFrenetSerret_),
      normalRadius(normalRadius_),
      binormalRadius(binormalRadius_),
      numPoints(numPoints_)
{
#ifdef PLB_DEBUG
    T eps = getEpsilon<T>(this->getPrecision());
#endif
    PLB_ASSERT(normalRadius > eps);
    PLB_ASSERT(binormalRadius > eps);
    if (numPoints < 3) {
        numPoints = 3;
    }
}

template<typename T, template<typename U> class Descriptor>
TubeStreamlineParticle3D<T,Descriptor>* TubeStreamlineParticle3D<T,Descriptor>::clone() const
{
    return new TubeStreamlineParticle3D<T,Descriptor>(*this);
}

template<typename T, template<typename U> class Descriptor>
int TubeStreamlineParticle3D<T,Descriptor>::getId() const 
{
    return id;
}

template<typename T, template<typename U> class Descriptor>
void TubeStreamlineParticle3D<T,Descriptor>::serialize(HierarchicSerializer& serializer) const
{
    StreamlineParticle3D<T,Descriptor>::serialize(serializer);
    serializer.addValue(normalRadius);
    serializer.addValue(binormalRadius);
    serializer.addValue(numPoints);
}

template<typename T, template<typename U> class Descriptor>
void TubeStreamlineParticle3D<T,Descriptor>::unserialize(HierarchicUnserializer& unserializer)
{
    StreamlineParticle3D<T,Descriptor>::unserialize(unserializer);
    unserializer.readValue(normalRadius);
    unserializer.readValue(binormalRadius);
    unserializer.readValue(numPoints);
}

template<typename T, template<typename U> class Descriptor>
void TubeStreamlineParticle3D<T,Descriptor>::rescale(int dxScale, int dtScale)
{
    StreamlineParticle3D<T,Descriptor>::rescale(dxScale, dtScale);

    T posDimDx = 1.0;
    T posDimDt = 0.0;
    T posScaleFactor = scaleFromReference(dxScale, posDimDx, dtScale, posDimDt);
    normalRadius *= posScaleFactor;
    binormalRadius *= posScaleFactor;
}

template<typename T, template<typename U> class Descriptor>
void TubeStreamlineParticle3D<T,Descriptor>::createStreamline(TriangleSet<T>& streamline) const
{
    static T pi = std::acos((T) -1);
    std::vector<Array<T,3> > const& positions = this->getPositions();

    streamline.clear();
    Precision precision = this->getPrecision();
    streamline.setPrecision(precision);
    if (positions.size() < 2) {
        return;
    }

    std::vector<T> t(positions.size());
    for (plint i = 0; i < (plint) positions.size(); i++) {
        t[i] = (T) i;
    }

    NaturalCubicSpline<T> *x, *y, *z;
    this->constructSplineVectorRepresentation(t, positions, &x, &y, &z);

    T dangle = 2.0 * pi / (T) numPoints;
    //T xi = 0.0;

    std::vector<Array<T,3> > tangents;
    std::vector<Array<T,3> > normals;
    std::vector<Array<T,3> > binormals;
    if (this->constructOrthonormalFrames(t, x, y, z, tangents, normals, binormals) != 0) {
        return;
    }

    std::vector<Array<T,3> > from(numPoints+1);
    std::vector<Array<T,3> > to(numPoints+1);
    for (plint iPoint = 0; iPoint <= numPoints; iPoint++) {
        T angle = (T) iPoint * dangle;
        T eta  = normalRadius * std::cos(angle);
        T zeta = binormalRadius * std::sin(angle);
        //Array<T,3> position = positions[0] + xi * tangents[0] + eta * normals[0] + zeta * binormals[0];
        Array<T,3> position = positions[0] + eta * normals[0] + zeta * binormals[0];
        from[iPoint] = position;
    }
    T signNormal = (T) 1;
    T signBinormal = (T) 1;
    for (plint i = 1; i < (plint) positions.size(); i++) {
        if (dot<T,3>(normals[i-1], normals[i]) < (T) 0) {
            signNormal = -signNormal;
        }
        if (dot<T,3>(binormals[i-1], binormals[i]) < (T) 0) {
            signBinormal = -signBinormal;
        }
        for (plint iPoint = 0; iPoint <= numPoints; iPoint++) {
            T angle = (T) iPoint * dangle;
            T eta  = normalRadius * std::cos(angle);
            T zeta = binormalRadius * std::sin(angle);
            //Array<T,3> position = positions[i] + xi * tangents[i] + eta * signNormal * normals[i] + zeta * signBinormal * binormals[i];
            Array<T,3> position = positions[i] + eta * signNormal * normals[i] + zeta * signBinormal * binormals[i];
            to[iPoint] = position;
        }
        TriangleSet<T> strip = constructStrip<T>(from, to);
        streamline.append(strip);
        from = to;
    }

    this->destructSplineVectorRepresentation(x, y, z);
}


/* *************** class StripStreamlineParticle3D ************************************ */

template<typename T, template<typename U> class Descriptor>
int StripStreamlineParticle3D<T,Descriptor>::id =
        meta::registerGenericParticle3D<T,Descriptor,StripStreamlineParticle3D<T,Descriptor> >("StripStreamline");

template<typename T, template<typename U> class Descriptor>
StripStreamlineParticle3D<T,Descriptor>::StripStreamlineParticle3D()
    : StreamlineParticle3D<T,Descriptor>(),
      thickness((T) 0)
{ }

template<typename T, template<typename U> class Descriptor>
StripStreamlineParticle3D<T,Descriptor>::StripStreamlineParticle3D(plint tag_, Array<T,3> const& position_,
        Array<T,3> const& velocity_, Box3D const& domain_, T thickness_, bool useFrenetSerret_)
    : StreamlineParticle3D<T,Descriptor>(tag_, position_, velocity_, domain_, useFrenetSerret_),
      thickness(thickness_)
{
    T eps = getEpsilon<T>(this->getPrecision());
    PLB_ASSERT(thickness > eps);
}

template<typename T, template<typename U> class Descriptor>
StripStreamlineParticle3D<T,Descriptor>::StripStreamlineParticle3D(plint tag_, Array<T,3> const& position_,
        Array<T,3> const& velocity_, bool useConstantSpatialStepIntegration_, T step_, Box3D const& domain_,
        T thickness_, bool useFrenetSerret_)
    : StreamlineParticle3D<T,Descriptor>(tag_, position_, velocity_, useConstantSpatialStepIntegration_,
            step_, domain_, useFrenetSerret_),
      thickness(thickness_)
{
#ifdef PLB_DEBUG
    T eps = getEpsilon<T>(this->getPrecision());
#endif
    PLB_ASSERT(thickness > eps);
}

template<typename T, template<typename U> class Descriptor>
StripStreamlineParticle3D<T,Descriptor>* StripStreamlineParticle3D<T,Descriptor>::clone() const
{
    return new StripStreamlineParticle3D<T,Descriptor>(*this);
}

template<typename T, template<typename U> class Descriptor>
int StripStreamlineParticle3D<T,Descriptor>::getId() const 
{
    return id;
}

template<typename T, template<typename U> class Descriptor>
void StripStreamlineParticle3D<T,Descriptor>::serialize(HierarchicSerializer& serializer) const
{
    StreamlineParticle3D<T,Descriptor>::serialize(serializer);
    serializer.addValue(thickness);
}

template<typename T, template<typename U> class Descriptor>
void StripStreamlineParticle3D<T,Descriptor>::unserialize(HierarchicUnserializer& unserializer)
{
    StreamlineParticle3D<T,Descriptor>::unserialize(unserializer);
    unserializer.readValue(thickness);
}

template<typename T, template<typename U> class Descriptor>
void StripStreamlineParticle3D<T,Descriptor>::rescale(int dxScale, int dtScale)
{
    StreamlineParticle3D<T,Descriptor>::rescale(dxScale, dtScale);

    T posDimDx = 1.0;
    T posDimDt = 0.0;
    T posScaleFactor = scaleFromReference(dxScale, posDimDx, dtScale, posDimDt);
    thickness *= posScaleFactor;
}

template<typename T, template<typename U> class Descriptor>
void StripStreamlineParticle3D<T,Descriptor>::createStreamline(TriangleSet<T>& streamline) const
{
    static T invSqrt2 = 1.0 / std::sqrt((T) 2);
    std::vector<Array<T,3> > const& positions = this->getPositions();

    streamline.clear();
    Precision precision = this->getPrecision();
    streamline.setPrecision(precision);
    if (positions.size() < 2) {
        return;
    }

    std::vector<T> t(positions.size());
    for (plint i = 0; i < (plint) positions.size(); i++) {
        t[i] = (T) i;
    }

    NaturalCubicSpline<T> *x, *y, *z;
    this->constructSplineVectorRepresentation(t, positions, &x, &y, &z);

    std::vector<Array<T,3> > tangents;
    std::vector<Array<T,3> > normals;
    std::vector<Array<T,3> > binormals;
    if (this->constructOrthonormalFrames(t, x, y, z, tangents, normals, binormals) != 0) {
        return;
    }

    T halfThickness = 0.5 * thickness;
    std::vector<Array<T,3> > from(positions.size());
    std::vector<Array<T,3> > to(positions.size());
    T signNormal = (T) 1;
    T signBinormal = (T) 1;
    for (plint i = 0; i < (plint) positions.size(); i++) {
        Array<T,3> direction;
        Array<T,3> position;
        if (i != 0) {
            if (dot<T,3>(normals[i-1], normals[i]) < (T) 0) {
                signNormal = -signNormal;
            }
            if (dot<T,3>(binormals[i-1], binormals[i]) < (T) 0) {
                signBinormal = -signBinormal;
            }
        }
        direction = signNormal * invSqrt2 * normals[i] + signBinormal * invSqrt2 * binormals[i];
        position = positions[i] + halfThickness * direction;
        from[i] = position;
        position = positions[i] - halfThickness * direction;
        to[i] = position;
    }

    this->destructSplineVectorRepresentation(x, y, z);
    streamline.append(constructStrip<T>(from, to));
}

}  // namespace plb

#endif  // STREAMLINE_PARTICLE_3D_H
