/* This file is part of the Palabos library.
 *
 * Copyright (C) 2011-2015 FlowKit Sarl
 * Route d'Oron 2
 * 1010 Lausanne, Switzerland
 * E-mail contact: contact@flowkit.com
 *
 * The most recent release of Palabos can be downloaded at 
 * <http://www.palabos.org/>
 *
 * The library Palabos is free software: you can redistribute it and/or
 * modify it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * The library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef TRIANGLE_HASHING_HH
#define TRIANGLE_HASHING_HH

#include "core/globalDefs.h"
#include "geometry/triangleHashing.h"
#include "atomicBlock/reductiveDataProcessingFunctional3D.h"
#include "atomicBlock/atomicContainerBlock3D.h"
#include "multiBlock/multiDataProcessorWrapper3D.h"
#include <algorithm>

namespace plb {

struct HashingData : public ContainerBlockData {
    HashingData (
            plint nx, plint ny, plint nz,
            Dot3D const& location )
        : triangles(nx,ny,nz)
    {
        triangles.setLocation(location);
    }
    virtual HashingData* clone() const {
        return new HashingData(*this);
    }
    ScalarField3D<std::vector<plint> > triangles;
    std::vector<Dot3D> assignedPositions;
};


/* ******** class TriangleHashing ********************************************* */

template<typename T>
TriangleHashing<T>::TriangleHashing(AtomicContainerBlock3D& hashContainer)
    : triangles (
        dynamic_cast<HashingData*>(hashContainer.getData())->triangles ),
      assignedPositions (
        dynamic_cast<HashingData*>(hashContainer.getData())->assignedPositions )
{ }

template<typename T>
void TriangleHashing<T>::getTriangles (
                Array<T,2> const& xRange,
                Array<T,2> const& yRange,
                Array<T,2> const& zRange,
                std::vector<plint>& foundTriangles ) const
{
    // Fit onto the grid by making it bigger, to be sure the triangle
    //   is never missed through round-off errors.
    Box3D discreteRange (
            (plint)xRange[0], (plint)xRange[1]+1,
            (plint)yRange[0], (plint)yRange[1]+1,
            (plint)zRange[0], (plint)zRange[1]+1 );
    getTriangles(discreteRange, foundTriangles);
}

template<typename T>
void TriangleHashing<T>::getTriangles (
        Box3D const& domain,
        std::vector<plint>& foundTriangles ) const
{
    Dot3D location(triangles.getLocation());
    // Convert to local coordinates.
    Box3D shifted(domain.shift(-location.x,-location.y,-location.z));
    foundTriangles.clear();
    Box3D inters;
    if (intersect(shifted, triangles.getBoundingBox(), inters)) {
        for (plint iX=inters.x0; iX<=inters.x1; ++iX) {
            for (plint iY=inters.y0; iY<=inters.y1; ++iY) {
                for (plint iZ=inters.z0; iZ<=inters.z1; ++iZ) {
                    std::vector<plint> const& newTriangles = triangles.get(iX,iY,iZ);
                    foundTriangles.insert(foundTriangles.end(),
                                          newTriangles.begin(), newTriangles.end());
                }
            }
        }
        std::sort(foundTriangles.begin(), foundTriangles.end());
        foundTriangles.erase( unique(foundTriangles.begin(), foundTriangles.end()),
                              foundTriangles.end() );
    }
}

template<typename T>
void TriangleHashing<T>::assignTriangles (
        ConnectedTriangleMesh<T>& mesh )
{
    typedef typename ConnectedTriangleMesh<T>::PTriangleIterator PTriangleIterator;
    typedef typename ConnectedTriangleMesh<T>::PTriangle PTriangle;

    plint triangleIDtagging = mesh.getTriangleTag("UniqueID");
    Dot3D location(triangles.getLocation());
    assignedPositions.clear();
    PTriangleIterator it = mesh.triangleIterator();
    while (!it->end()) {
        PTriangle triangle = it->next();
        Array<T,3> const& vertex0 = triangle->vertex(0)->get();
        Array<T,3> const& vertex1 = triangle->vertex(1)->get();
        Array<T,3> const& vertex2 = triangle->vertex(2)->get();

        Array<T,2> xRange (
                     std::min(vertex0[0], std::min(vertex1[0], vertex2[0])),
                     std::max(vertex0[0], std::max(vertex1[0], vertex2[0])) );
        Array<T,2> yRange (
                     std::min(vertex0[1], std::min(vertex1[1], vertex2[1])),
                     std::max(vertex0[1], std::max(vertex1[1], vertex2[1])) );
        Array<T,2> zRange (
                     std::min(vertex0[2], std::min(vertex1[2], vertex2[2])),
                     std::max(vertex0[2], std::max(vertex1[2], vertex2[2])) );

        // Fit onto the grid by making it bigger, to be sure the triangle
        //   is never missed through round-off errors.
        Box3D discreteRange (
                (plint)xRange[0], (plint)xRange[1]+1,
                (plint)yRange[0], (plint)yRange[1]+1,
                (plint)zRange[0], (plint)zRange[1]+1 );
        // Convert to local coordinates.
        discreteRange = discreteRange.shift (
                -location.x, -location.y, -location.z );
        Box3D inters;
        if (intersect(discreteRange, triangles.getBoundingBox(), inters))
        {
            for (plint iX=inters.x0; iX<=inters.x1; ++iX) {
                for (plint iY=inters.y0; iY<=inters.y1; ++iY) {
                    for (plint iZ=inters.z0; iZ<=inters.z1; ++iZ) {
                        if (triangles.get(iX,iY,iZ).empty()) {
                            assignedPositions.push_back(Dot3D(iX,iY,iZ));
                        }
                        triangles.get(iX,iY,iZ).push_back(triangle->tag(triangleIDtagging));
                    }
                }
            }
        }
    }
}


/* ******** MeshToHash ************************************ */

template<typename T>
MeshToHash<T>::MeshToHash (
        ConnectedTriangleMesh<T>& mesh_ )
    :  mesh(mesh_)
{ }

template<typename T>
void MeshToHash<T>::processGenericBlocks (
        Box3D domain, std::vector<AtomicBlock3D*> blocks )
{
    PLB_PRECONDITION( blocks.size()==1 );
    AtomicContainerBlock3D* container =
        dynamic_cast<AtomicContainerBlock3D*>(blocks[0]);
    PLB_ASSERT( container );
    HashingData* hashData
        = new HashingData (
                container->getNx(), container->getNy(), container->getNz(),
                container->getLocation() );
    container->setData(hashData);
    TriangleHashing<T>(*container).assignTriangles(mesh);
}

template<typename T>
MeshToHash<T>* MeshToHash<T>::clone() const {
    return new MeshToHash<T>(*this);
}

template<typename T>
void MeshToHash<T>::getTypeOfModification(std::vector<modif::ModifT>& modified) const {
    modified[0] = modif::staticVariables;  // Container Block with hash data.
}

template<typename T>
BlockDomain::DomainT MeshToHash<T>::appliesTo() const {
    return BlockDomain::bulk;
}


template<typename T>
void setupHashing(MultiContainerBlock3D& container, ConnectedTriangleMesh<T>& mesh)
{
    std::vector<MultiBlock3D*> hashArg;
    hashArg.push_back(&container);
    applyProcessingFunctional (
            new MeshToHash<T>(mesh),
            container.getBoundingBox(), hashArg );
}

}  // namespace plb

#endif  // TRIANGLE_HASHING_HH

